from dataclasses import dataclass
from enum import Enum
from functools import cached_property
from typing import (
    Annotated,
    Any,
    Callable,
    List,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import numpy as np
from annotated_types import BaseMetadata, Gt, Len, Lt, Predicate
from psygnal import evented
from pydantic import BaseModel, Field, field_validator, model_validator


@dataclass(frozen=True, slots=True)
class Step(BaseMetadata):
    """Gt(gt=x) implies that the value must be greater than x.

    It can be used with any type that supports the ``>`` operator,
    including numbers, dates and times, strings, sets, and so on.
    """

    x: int


class mode(str, Enum):
    GAUSSIAN = "GAUSSIAN"
    DONUT = "DONUT"
    BOTTLE = "BOTTLE"
    DONUT_BOTTLE = "DONUT BOTTLE"
    LOADED = "LOADED"


class window(str, Enum):
    NO = "NO"
    OLD = "OLD"
    NEW = "NEW"
    CUSTOM = "CUSTOM"


class normalize(str, Enum):
    YES = "YES"
    NO = "NO"


class noise(str, Enum):
    YES = "YES"
    NO = "NO"


class polarization(int, Enum):
    ELLIPTICAL = 1
    RADIAL = 2
    AZIMUTHAL = 3


class AberrationFloat(float):
    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def validate(cls, v, info):
        if not isinstance(v, float):
            raise TypeError("Float required")
        # you could also return a string here which would mean model.post_code
        # would be a string, pydantic won't care but you could end up with some
        # confusion since the value's type won't match the type annotation
        # exactly
        return v


@evented
class PSFConfig(BaseModel):
    # sampling parameters
    L_obs_XY: float = Field(default=2, description="Observation scale in XY (in µm)")
    L_obs_Z: float = Field(default=2, description="Observation scale in Z (in µm)")
    Nxy: int = Field(
        default=31,
        description="Discretization of image plane - better be odd number for perfect 0",
    )
    Nz: int = Field(
        default=31,
        description="Discretization of Z axis - better be odd number for perfect 0",
    )
    Ntheta: int = Field(
        default=31, description="Integration sted of the focalization angle"
    )
    Nphi: int = Field(
        default=31, description="Integration sted of the aximutal angle on the pupil"
    )

    # Normalization
    Normalize: normalize = (
        normalize.YES
    )  # Normalize the intensity of the PSF to have a maximum of 1

    # Geometry parameters
    NA: float = Field(default=1, description="Numerical Aperture of Objective Lens")
    WD: float = Field(
        default=2800, description="Working Distance of the objective lens (in µm)"
    )
    n1: float = Field(
        default=1.33, description="Refractive index of the immersion medium"
    )
    n2: float = Field(default=1.52, description="Refractive index of the coverslip")
    n3: float = Field(default=1.38, description="Refractive index of the sample")
    Thickness: float = Field(
        default=170, description="Thickness of the coverslip (in µm)"
    )
    Collar: float = Field(
        default=170,
        description="Correction collar setting to compensate coverslip thickness",
        gt=0,
        lt=300,
    )
    Depth: float = Field(
        default=0, description="Imaging depth in the sample (in µm)", ge=0, lt=150
    )
    Tilt: float = Field(
        default=0, description="Tilt angle of the coverslip (in °)", ge=-10, lt=10
    )

    Window: window = window.NO

    Wind_Radius: float = Field(
        default=2.3, description="Diameter of the cranial window (in mm)"
    )
    Wind_Depth: float = Field(
        default=2.23, description="Depth of the cranial window (in mm)"
    )

    # Aberrations
    a0: AberrationFloat = Field(default=0, description="Piston", gt=-1, lt=1)
    a1: AberrationFloat = Field(default=0, description="Vertical Tilt", gt=-1, lt=1)
    a2: AberrationFloat = Field(default=0, description="Horizontal Tilt", gt=-1, lt=1)
    a3: AberrationFloat = Field(
        default=0, description="Oblique Astigmatism", gt=-1, lt=1
    )
    a4: AberrationFloat = Field(default=0, description="Defocus", gt=-1, lt=1)
    a5: AberrationFloat = Field(
        default=0, description="Vertical Astigmatism", gt=-1, lt=1
    )
    a6: AberrationFloat = Field(default=0, description="Vertical Trefoil", gt=-1, lt=1)
    a7: AberrationFloat = Field(default=0, description="Vertical Coma", gt=-1, lt=1)
    a8: AberrationFloat = Field(default=0, description="Horizontal Coma", gt=-1, lt=1)
    a9: AberrationFloat = Field(default=0, description="Oblique Trefoil", gt=-1, lt=1)
    a12: AberrationFloat = Field(
        default=0, description="Primary spherical", gt=-1, lt=1
    )
    Aberration_offset_x: float = Field(
        default=0,
        description="X offset of the aberration function in regard to pupil center",
    )
    Aberration_offset_y: float = Field(
        default=0,
        description="Y offset of the aberration function in regard to pupil center",
    )

    # Beam parameters
    Mode: mode = mode.GAUSSIAN
    Polarization: polarization = polarization.ELLIPTICAL
    Wavelength: float = Field(default=0.592, description="Wavelength of light (in µm)")
    Waist: float = Field(
        default=8000,
        description="Diameter of the input beam on the objective pupil (in µm)",
        gt=0,
        lt=20000,
    )
    Ampl_offset_x: float = Field(
        default=0,
        description="X offset of the amplitude profile in regard to pupil center",
    )
    Ampl_offset_y: float = Field(
        default=0,
        description="Y offset of the amplitude profile in regard to pupil center",
    )

    # Polarization parameters
    Psi: float = Field(
        default=0, description="Direction of the polarization (in °)", ge=0, le=180
    )
    Epsilon: float = Field(
        default=45, description="Ellipticity of the polarization (in °)", ge=-45, le=45
    )

    # STED parameters
    VC: float = Field(
        default=1.0,
        description="Vortex charge (should be integer to produce donut)",
        gt=-6,
        lt=6,
    )
    RC: float = Field(
        default=1.0,
        description="Ring charge (should be odd to produce bottle)",
        gt=-5,
        lt=5,
    )
    Ring_Radius: float = Field(
        default=0.707,
        description="Radius of the ring phase mask (on unit pupil)",
        gt=0,
        lt=1,
    )
    Mask_offset_x: float = Field(
        default=0, description="X offset of the phase mask in regard to pupil center"
    )
    Mask_offset_y: float = Field(
        default=0, description="Y offset of the phase mask in regard to pupil center"
    )
    p: float = Field(
        default=0.5,
        description="Ratio between Donut (p) and Bottle (1-p) intensity",
        gt=0,
        lt=1,
    )

    loaded_phase_mask: Optional[np.ndarray] = Field(
        default=None, description="Loaded Phasemak"
    )

    # Noise Parameters
    Add_noise: noise = noise.YES  # Add noise to the PSF

    Gaussian_beam_noise: Annotated[float, Step(0.1)] = Field(
        default=0.0, description="Gaussian_beam noise", ge=0, lt=1
    )
    Detector_gaussian_noise: Annotated[float, Step(0.1)] = Field(
        default=0.0, description="Detector Gaussian noise", ge=0, lt=1
    )
    # Add_detector_poisson_noise: bool = False  # standard deviation of the noise

    @property
    def t_wind(self):
        if self.Window == window.CUSTOM:
            return self.Wind_Depth * 1e3
        else:
            return 2.23e3

    @property
    def r_wind(self):  # No cranial window
        if self.Window == window.NO:
            return 100 * self.t_wind
        elif self.Window == window.OLD:  # Old cylindrical cranial window
            return 1.5e3
        elif self.Window == window.NEW:  # New conical cranial window
            return 2.3e3
        elif self.Window == window.CUSTOM:
            return self.Wind_Radius * 1e3
        else:
            raise NotImplementedError("Please use a specified window type")

    @property
    def k0(self):  # Wavevector (in µm^-1)
        return 2 * np.pi / self.Wavelength

    @property
    def alpha(self):  # maximum focusing angle of the objective (in rad)
        return np.arcsin(self.NA / self.n1)

    @property
    def r0(self):  # radius of the pupil (in µm)
        return self.WD * np.sin(self.alpha)

    # convert angle in red
    @property
    def gamma(self):  # tilt angle (in rad)
        return self.Tilt * np.pi / 180

    @property
    def psi(self):  # polar direction (in rad)
        return self.Psi * np.pi / 180

    @property
    def eps(self):  # ellipticity (in rad)
        return self.Epsilon * np.pi / 180

    @property
    def sg(self):
        return np.sin(self.gamma)

    @property
    def cg(self):
        return np.cos(self.gamma)

    @property
    def alpha_eff(
        self,
    ):  # effective focalization angle due to the cranial window (in rad)
        return min(np.arctan(self.r_wind / self.t_wind), self.alpha)

    @property
    def NA_eff(self):  # Effective NA in presence of the cranial window
        return min(self.n1 * np.sin(self.alpha_eff), self.NA)

    @property
    def r0_eff(self):  # Effective pupil radius (in µm)
        return self.WD * np.sin(self.alpha_eff)

    @property
    def alpha_int(self):  # Integration range (in rad)
        return self.alpha_eff + abs(self.gamma)

    @property
    def r0_int(self):  # integration radius on pupil (in µm)
        return self.WD * np.sin(self.alpha_int)

    @property
    def alpha2_eff(self):
        return np.arcsin((self.n1 / self.n2) * np.sin(self.alpha_eff))

    @property
    def alpha3_eff(self):
        return np.arcsin((self.n2 / self.n3) * np.sin(self.alpha2_eff))

    @property
    def Dfoc(self):  # Corrected focus position
        # return 0.0564 * self.Depth + 0.1692 * (self.Thickness - self.Collar)  # No aberration correction
        return (
            1
            / np.tan(self.alpha3_eff)
            * (
                self.Depth * (np.tan(self.alpha_eff) - np.tan(self.alpha3_eff))
                + (self.Thickness - self.Collar)
                * (np.tan(self.alpha_eff) - np.tan(self.alpha2_eff))
            )
        )

    @property
    def deltatheta(self):
        return self.alpha_int / self.Ntheta

    @property
    def deltaphi(self):
        return 2 * np.pi / self.Nphi

    @model_validator(mode="after")
    def validate_NA(cls, self: "PSFConfig", info):
        NA = self.NA
        if NA <= 0:
            raise ValueError("numerical_aperature must be positive")
        if self.n1 < NA:
            raise ValueError(
                "numerical_aperature must be smaller than the refractive index"
            )
        if self.Mode == mode.LOADED:
            if self.loaded_phase_mask is None:
                raise ValueError("You need to load a phase mask to use the loaded mode")

        return self

    class Config:
        validate_assignment = True
        extra = "forbid"
        arbitrary_types_allowed = True


PSFGenerator = Callable[[PSFConfig], np.ndarray]



import json

import rich_click as click
import tifffile

from faser.generators.base import PSFConfig
from faser.generators.vectorial.stephane.tilted_coverslip import generate_psf

arg_map = {}

logo = r"""
  __                     
 / _| __ _ ___  ___ _ __ 
| |_ / _` / __|/ _ \ '__|
|  _| (_| \__ \  __/ |   
|_|  \__,_|___/\___|_|   
"""


def make_nice_name(name):
    return name.replace("_", "_").lower()


def psf_attributes(func):

    for key, field in PSFConfig.model_fields.items():

        nice_key = make_nice_name(key)

        arg_map[nice_key] = key

        func = click.option(
            f"--{nice_key}",
            type=field.annotation,
            help=field.description,
            default=field.default,
        )(func)
    return func


@click.command()
@click.option(
    "--config", type=click.File(mode="r"), help="Path to a JSON file", default=None
)
@psf_attributes
def main(config=None, **kwargs):

    click.echo(logo)

    # remape
    kwargs = {arg_map[k]: v for k, v in kwargs.items()}

    if config is not None:
        click.echo(f"Loading config from `{config.name}`")
        kwargs = {
            **json.load(config),
            **kwargs,
        }

    config = PSFConfig(**kwargs)

    click.echo(f"Generating PSF with config")
    x = generate_psf(config)

    tifffile.imsave("psf.tif", x)

    with open("psf_config.json", "w") as f:
        f.write(config.model_dump_json(indent=2))



from scipy import ndimage
import numpy as np
from faser.generators.base import PSFConfig, Mode
from skimage import data
from faser.generators.vectorial.stephane import generate_psf

#%%
config = PSFConfig(Nx=32, Ny=32, Nz=32, mode=Mode.GAUSSIAN)

x = np.expand_dims(data.coins(), axis=2)
print(x.shape)
h1 = generate_psf(config)
print("Here")

print(ndimage.convolve(x, h1, mode="constant", cval=0.0, origin=0))
print



import os


FASER_PATH = os.path.dirname(os.path.realpath(__file__))
ASSETS_PATH = os.path.join(FASER_PATH, "assets")


def get_asset_file(file, darkMode=False):
    return os.path.join(ASSETS_PATH, file)



from faser.generators.base import PSFGenerator
from faser.generators.scalar.gibson_lanny.gen import generate_psf


class GibsonLannyPSFGenerator(PSFGenerator):
    def generate(self):
        return generate_psf(self.config)



from faser.napari.widgets.main_widget import FloatRange, IntRange
from pydantic import BaseModel
from typing import List
import itertools
import dask
import time
import dask.array as da
import numpy as np

# overwrite default with multiprocessing scheduler


class Model(BaseModel):
    x: float
    y: float
    z: int
    options: str
    f: int


mapper = {
    "x": FloatRange(min=0, max=1, steps=2),
    "y": FloatRange(min=0, max=1, steps=3),
    "z": IntRange(min=0, max=1, steps=3),
    "options": OptionRange(options=["a", "b"]),
}


@dask.delayed
def process_map(model_values):
    return np.zeros((10, 10, 10), dtype=np.float64)


def start_map():

    model = Model(x=0, y=0, z=0, options="a", f=5)

    permutations = list(itertools.product(*map(lambda x: x.to_list(), mapper.values())))

    models = []

    for i in permutations:
        values = {k: v for k, v in zip(mapper.keys(), i)}
        new_model = model.copy(update=values)
        models.append(new_model)
        print(new_model)

    t = [
        da.from_delayed(process_map(new_model.dict()), (10, 10, 10), dtype=np.float64)
        for new_model in models
    ]
    t = da.stack(t)

    print(t.shape)

    print(t.compute())

    return mapper


def split(l, n):
    return [
        da.stack(l[int(i * len(l) / n) : int((i + 1) * len(l) / n - 1)])
        for i in range(n)
    ]


def new_map():

    model = Model(x=0, y=0, z=0, options="a", f=5)

    permutations = list(itertools.product(*map(lambda x: x.to_list(), mapper.values())))

    models = []

    for i in permutations:
        values = {k: v for k, v in zip(mapper.keys(), i)}
        new_model = model.copy(update=values)
        models.append(new_model)
        print(new_model)
        arrays = da.from_delayed(
            process_map(new_model.dict()), shape=(10, 10, 10), dtype=np.float64
        )

    x = arrays

    print("split Arrays")
    for i in reversed(mapper.values()):
        x = len(i.to_list())
        print(arrays)
        arrays = split(arrays, x)

    print(arrays)

    return mapper


if __name__ == "__main__":
    new_map()



#%%
from faser.generators.base import PSFGeneratorConfig
from faser.generators.scalar.phasenet import PhaseNetPSFGenerator
from faser.generators.vectorial.stephane import StephanePSFGenerator

import matplotlib.pyplot as plt
import numpy as np
import matplotlib
import napari
from faser.io import load
from prysm.psf import centroid
from scipy.ndimage import center_of_mass
from faser.retrievers.gs import ger_sax

config = PSFGeneratorConfig()


x = load("test_phase_net.zarr")
s = load("test_stephane.zarr")

xn = x / np.max(x)
sn = s / np.max(s)


#%%
print(center_of_mass(xn), center_of_mass(sn))

pm = ger_sax(xn[xn.shape[0] // 2], 30)
print(pm.shape)

print(x.shape, s.shape)


#%%
viewer = napari.Viewer()
viewer.add_image(xn)
viewer.add_image(sn)
viewer.add_image(pm)
viewer.add_image(np.abs(xn - sn))
napari.run()
# %%



import json
from typing import Callable
import numpy as np


def save(generator: Callable, config, path, as_zarr=False, **kwargs):
    """Save a generator to a zarr file.

    Parameters
    ----------
    generator : generator
        The generator to save.
    path : str
        The path to the zarr file.
    **kwargs
        Additional keyword arguments to pass to xarray.to_zarr.

    """

    data = generator(config).generate()

    if as_zarr:
        with open(path, "wb") as f:
            np.save(f, data)
        return path

    else:
        import xarray as xr
        import zarr as zr

        store = zr.DirectoryStore(path, compression=None)

        return (
            xr.DataArray(
                data,
                dims=list("xyz"),
                attrs={"config": config.json(), "generator": generator.__name__},
            )
            .to_dataset(name="data")
            .to_zarr(store, mode="w", compute=True)
        )


def load(
    path,
    as_zarr=False,
):
    if not as_zarr:
        with open(path, "rb") as f:
            data = np.load(f)
        return data

    import xarray as xr

    return xr.open_zarr(path)["data"]



# import itertools
# import os
import typing
from enum import Enum
from typing import Any, Callable, List, Type

# import dask
# import dask.array as da
import napari
import numpy as np
import pydantic
from annotated_types import Ge, Gt, Le, Lt

# from faser.generators.base import AberrationFloat, PSFConfig
# from faser.generators.vectorial.stephane.tilted_coverslip import generate_psf
from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo
from PyQt5 import QtCore
from PyQt5.QtWidgets import QWidget
from qtpy import QtGui, QtWidgets
from superqt import QEnumComboBox  # QDoubleRangeSlider,
from superqt import QLabeledDoubleRangeSlider, QLabeledDoubleSlider

from faser.env import get_asset_file

batch_png = QtGui.QPixmap(get_asset_file("batch.png"))
single_png = QtGui.QPixmap(get_asset_file("single.png"))


def get_field_gt(field: FieldInfo) -> float:
    for metadata in field.metadata:
        if isinstance(metadata, Gt):
            return metadata.gt
        if isinstance(metadata, Ge):
            return metadata.ge


def get_field_lt(field: FieldInfo) -> float:
    for metadata in field.metadata:
        if isinstance(metadata, Lt):
            return metadata.lt
        if isinstance(metadata, Le):
            return metadata.le


class FormField(QtWidgets.QWidget):
    on_child_value_changed = QtCore.pyqtSignal(str, object)
    on_child_range_value_changed = QtCore.pyqtSignal(str, object)

    def __init__(
        self, key: str, field: FieldInfo, toggable: bool = True, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.field = field
        self.key = key
        self.label = key
        self.child = None
        self.mode = "single"
        self.description = field.description
        self.steps = 3
        self.toggable = toggable
        self.range_child = None

    def replace_widget(self, oldwidget, newwidget):
        self.xlayout.removeWidget(oldwidget)
        oldwidget.setParent(None)
        self.xlayout.addWidget(newwidget)

    def reset_value(self, value):
        raise NotImplementedError("reset_value() must be implemented")

    def reset_default(self):
        self.reset_value(self.field.default)

    def emit_child_value_changed(self, value):
        self.on_child_value_changed.emit(self.key, value)

    def on_child_range_changed(self, value):
        self.on_child_range_value_changed.emit(self.key, value)

    def on_change_mode(self):

        if self.mode == "single":
            self.toggle_button.setIcon(QtGui.QIcon(batch_png))
            self.toggle_button.setIconSize(QtCore.QSize(15, 15))
            # self.emit_child_value_changed(None)
            self.replace_widget(self.child, self.range_child)

        elif self.mode == "range":
            self.toggle_button.setIcon(QtGui.QIcon(single_png))
            self.toggle_button.setIconSize(QtCore.QSize(15, 15))
            # self.on_child_range_changed(None)
            self.replace_widget(self.range_child, self.child)

        self.mode = "range" if self.mode == "single" else "single"

    def init_ui(self):
        assert self.child is not None, "Child widget must be set before init_ui()"
        assert (
            self.range_child is not None
        ), "Range_child widget must be set before init_ui()"
        self.xlayout = QtWidgets.QVBoxLayout()
        self.xlayout.setContentsMargins(0, 0, 0, 0)
        self.xlayout.setSpacing(1)

        self.label = QtWidgets.QLabel(self.label)
        self.label.setToolTip(self.description or "No description yet")
        layout = QtWidgets.QHBoxLayout()
        self.labelWidget = QtWidgets.QWidget()

        if self.toggable:
            self.toggle_button = QtWidgets.QPushButton()
            self.toggle_button.setIcon(QtGui.QIcon(single_png))
            self.toggle_button.setIconSize(QtCore.QSize(15, 15))
            self.toggle_button.clicked.connect(self.on_change_mode)
            layout.addWidget(self.toggle_button)

        layout.addWidget(self.label)
        layout.addStretch()

        # if self.mode == "single":
        #     layout.addWidget(self.child)
        # elif self.mode == "range":
        #     layout.addWidget(self.range_child)

        self.labelWidget.setLayout(layout)
        self.xlayout.addWidget(self.labelWidget)
        self.xlayout.addWidget(self.child)

        self.setLayout(self.xlayout)
        # self.setLayout(layout)

    def init_ui_helper_eff(self):
        assert self.child is not None, "Child widget must be set before init_ui()"
        assert (
            self.range_child is not None
        ), "Range_child widget must be set before init_ui()"
        self.xlayout = QtWidgets.QVBoxLayout()
        self.xlayout.setContentsMargins(0, 0, 0, 0)
        self.xlayout.setSpacing(1)

        self.label = QtWidgets.QLabel(self.label)
        self.label.setToolTip(self.description or "No description yet")
        layout = QtWidgets.QHBoxLayout()
        self.labelWidget = QtWidgets.QWidget()

        if self.toggable:
            self.toggle_button = QtWidgets.QPushButton()
            self.toggle_button.setIcon(QtGui.QIcon(single_png))
            self.toggle_button.setIconSize(QtCore.QSize(15, 15))
            self.toggle_button.clicked.connect(self.on_change_mode)
            layout.addWidget(self.toggle_button)

        layout.addWidget(self.label)
        # layout.addStretch()

        # if self.mode == "single":
        #     layout.addWidget(self.child)
        # elif self.mode == "range":
        #     layout.addWidget(self.range_child)

        self.labelWidget.setLayout(layout)
        self.xlayout.addWidget(self.labelWidget)
        self.xlayout.addWidget(self.child)

        self.setLayout(self.xlayout)
        # self.setLayout(layout)

    def init_ui_helper(self):
        # self.xlayout = QtWidgets.QVBoxLayout()
        self.xlayout = QtWidgets.QHBoxLayout()
        self.xlayout.setContentsMargins(0, 0, 0, 0)
        self.xlayout.setSpacing(1)

        self.label = QtWidgets.QLabel(self.label)
        self.label.setToolTip(self.description or "No description yet")
        layout = QtWidgets.QHBoxLayout()
        self.labelWidget = QtWidgets.QWidget()

        layout.addWidget(self.label)
        # layout.addStretch()

        self.labelWidget.setLayout(layout)
        self.xlayout.addWidget(self.labelWidget)
        self.xlayout.addWidget(self.child)

        self.setLayout(self.xlayout)
        # self.setLayout(layout)


class FloatRange(pydantic.BaseModel):
    min: float
    max: float
    steps: int

    def to_list(self):
        return np.linspace(self.min, self.max, self.steps, dtype=np.float64).tolist()


class FloatRangeStepSliderField(QtWidgets.QWidget):
    on_range_changed = QtCore.pyqtSignal(object)

    def __init__(self, *args, gt=None, lt=None, steps=None, **kwargs):
        super().__init__(*args, **kwargs)
        # self.layout = QtWidgets.QHBoxLayout()
        self.xlayout = QtWidgets.QHBoxLayout()
        self.xlayout.setContentsMargins(0, 0, 0, 0)
        self.xlayout.setSpacing(1)

        self.range_slider = QLabeledDoubleRangeSlider(QtCore.Qt.Horizontal)
        self.range_slider.setRange(gt or 0.0, lt or 1.0)
        self.range_slider.setValue((0.2, 0.8))
        self.range_slider.setStyleSheet("QWidget {font-size: 8pt;}")

        self.text_input = QtWidgets.QLineEdit()
        self.text_input.setFixedWidth(15)
        self.text_input.setText(str(steps) or 3)
        self.text_input.setValidator(QtGui.QIntValidator())

        self.range_slider.valueChanged.connect(self.on_range_valued_callback)
        # self.layout.addWidget(self.range_slider)
        # self.layout.addWidget(self.text_input)
        self.xlayout.addWidget(self.range_slider)
        self.xlayout.addSpacing(25)
        self.xlayout.addWidget(self.text_input)
        self.text_input.textChanged.connect(self.on_text_changed)

        # self.setLayout(self.layout)
        self.setLayout(self.xlayout)

    def on_range_valued_callback(self, value):
        self.on_range_changed.emit(
            FloatRange(
                min=float(value[0]),
                max=float(value[1]),
                steps=int(self.text_input.text()),
            )
        )

    def on_text_changed(self, value):
        self.on_range_changed.emit(
            FloatRange(
                min=float(self.range_slider.value()[0]),
                max=float(self.range_slider.value()[1]),
                steps=int(value),
            )
        )


class IntRange(pydantic.BaseModel):
    min: int
    max: int
    steps: int

    def to_list(self):
        return np.linspace(self.min, self.max, self.steps, dtype=np.int64).tolist()


class IntRangeStepSliderField(QtWidgets.QWidget):
    on_range_changed = QtCore.pyqtSignal(object)

    def __init__(self, *args, gt=None, lt=None, steps=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.layout = QtWidgets.QHBoxLayout()

        self.range_slider = QLabeledDoubleRangeSlider(QtCore.Qt.Horizontal)
        self.range_slider.setRange(gt or 0.0, lt or 1.0)
        self.text_input = QtWidgets.QLineEdit()
        self.text_input.setFixedWidth(20)
        self.text_input.setText(str(steps) or 3)
        self.text_input.setValidator(QtGui.QIntValidator())

        self.range_slider.valueChanged.connect(self.on_range_valued_callback)

        self.layout.addWidget(self.range_slider)
        self.layout.addWidget(self.text_input)
        self.text_input.textChanged.connect(self.on_text_changed)
        self.setLayout(self.layout)

    def on_range_valued_callback(self, value):
        self.on_range_changed.emit(
            IntRange(
                min=int(value[0]), max=int(value[1]), steps=int(self.text_input.text())
            )
        )

    def on_text_changed(self, value):
        self.on_range_changed.emit(
            IntRange(
                min=int(self.range_slider.value()[0]),
                max=int(self.range_slider.value()[1]),
                steps=int(value),
            )
        )


class FloatSliderField(FormField):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.child = QLabeledDoubleSlider(QtCore.Qt.Horizontal)
        self.child.setMinimum(get_field_gt(self.field) or 0.0)
        self.child.setValue(self.field.default)
        self.child.setMaximum(get_field_lt(self.field) or 1.0)
        self.child.valueChanged.connect(self.emit_child_value_changed)

        self.range_child = FloatRangeStepSliderField(
            gt=get_field_gt(self.field) or 0.0,
            lt=get_field_lt(self.field) or 1,
            steps=3,
        )
        self.range_child.on_range_changed.connect(self.on_child_range_changed)

    def reset_value(self, value):
        self.child.setValue(value)


class IntInputField(FormField):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.child = QtWidgets.QLineEdit()
        self.child.setText(str(self.field.default))
        self.child.setValidator(QtGui.QIntValidator())
        self.child.textChanged.connect(self.emit_text_changed)

        self.range_child = IntRangeStepSliderField(
            gt=get_field_gt(self.field) or 0.0,
            lt=get_field_lt(self.field) or 1,
            steps=3,
        )
        self.range_child.on_range_changed.connect(self.on_child_range_changed)

    def emit_text_changed(self, value):
        self.emit_child_value_changed(int(self.child.text()))

    def reset_value(self, value):
        self.child.setText(str(value))


class FloatInputField(FormField):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.child = QtWidgets.QLineEdit()
        self.child.setText(str(self.field.default))
        self.child.textChanged.connect(self.emit_text_changed)

        self.range_child = FloatRangeStepSliderField(
            gt=get_field_gt(self.field) or 0.0,
            lt=get_field_lt(self.field) or 1,
            steps=3,
        )
        self.range_child.on_range_changed.connect(self.on_child_range_changed)

    def emit_text_changed(self, value):
        self.emit_child_value_changed(float(self.child.text()))

    def reset_value(self, value):
        self.child.setText(str(value))


class OptionRange(pydantic.BaseModel):
    options: List[Any]

    def to_list(self):
        return self.options


class MultiEnumField(QtWidgets.QWidget):
    on_range_changed = QtCore.pyqtSignal(object)

    def __init__(self, *args, enum: Enum, **kwargs):
        super().__init__(*args, **kwargs)
        self.enum = enum
        self.layout = QtWidgets.QHBoxLayout()
        self.checkboxes = []
        self.checkable_values = []

        for i in enum:
            check = QtWidgets.QPushButton(i.name)
            check.setCheckable(True)
            check.clicked.connect(self.on_change_callback)

            self.checkboxes.append(check)
            self.checkable_values.append(i.value)
            self.layout.addWidget(check)

        self.setLayout(self.layout)

    def on_change_callback(self):
        print("Changed")
        check_enums = []

        for i, check in enumerate(self.checkboxes):
            if check.isChecked():
                check_enums.append(self.checkable_values[i])
            else:
                False

        self.on_range_changed.emit(OptionRange(options=check_enums))

    def reset_value(self, value):
        for i, check in enumerate(self.checkboxes):
            if self.checkable_values[i] in value:
                check.setChecked(True)
            else:
                check.setChecked(False)


class EnumField(FormField):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.child = QEnumComboBox()
        self.child.setEnumClass(self.field.annotation)
        self.child.currentEnumChanged.connect(self.emit_child_value_changed)

        self.range_child = MultiEnumField(enum=self.field.annotation)
        self.range_child.on_range_changed.connect(self.on_child_range_changed)
        # TODO: Implement

    def reset_value(self, value):
        self.child.setCurrentEnum(value)


def generate_single_widgets_from_model(
    model: Type[BaseModel],
    callback: Callable[[str, Any], None],
    range_callback: Callable[[str, Any], None] = None,
    parent=None,
    filter_fields=None,
) -> typing.List[FormField]:

    widgets = []

    for key, value in model.model_fields.items():

        if filter_fields:
            if not filter_fields(key, value):
                continue

        widget = None

        field_type = value.annotation
        print(field_type)

        if issubclass(field_type, float):
            widget = FloatSliderField(key, value, parent=parent)

        elif field_type == float:
            widget = FloatInputField(key, value, parent=parent)

        elif field_type == int:
            widget = IntInputField(key, value, parent=parent)

        elif issubclass(field_type, Enum):
            widget = EnumField(key, value, parent=parent)

        else:
            print(field_type)

        if widget is not None:
            if callback:
                widget.on_child_value_changed.connect(callback)
            if range_callback:
                widget.on_child_range_value_changed.connect(range_callback)
            widgets.append(widget)

    return widgets


def build_key_filter(
    field_set: typing.Set[str],
) -> typing.Callable[[str, FieldInfo], bool]:
    def key_filter(key, value):
        return key in field_set

    return key_filter



#%%
from faser.generators.base import Aberration, PSFGeneratorConfig, Mode
from faser.generators.scalar.phasenet import PhaseNetPSFGenerator
from faser.generators.vectorial.stephane import StephanePSFGenerator
import matplotlib.pyplot as plt
import numpy as np
import matplotlib
import napari

config = PSFGeneratorConfig()


x = PhaseNetPSFGenerator(config)
z = StephanePSFGenerator(config)

#%%
h1 = x.generate()


def gen(abs: Aberration):
    return StephanePSFGenerator(PSFGeneratorConfig(aberration=abs)).generate()


def gen_don(abs: Aberration):
    return StephanePSFGenerator(
        PSFGeneratorConfig(aberration=abs, mode=Mode.DONUT)
    ).generate()


x = {
    "Gaussian": {"Pure": gen(Aberration(a5=0)), "t": gen(Aberration(a5=0.7))},
    "Donut": {"Pure": gen_don(Aberration(a5=0)), "t": gen_don(Aberration(a5=0.7))},
}

nrows = len(x.keys())

#%%
fix, axes = plt.subplots(nrows=nrows, ncols=len(x["Gaussian"].keys()), figsize=(20, 20))


for i, row in enumerate(x.items()):
    key, value = row
    for y, col in enumerate(value.items()):
        subkey, image = col
        mid_plane = image.shape[0] // 2
        axes[i, y].imshow(image[mid_plane], cmap="hsv", interpolation="bilinear")


plt.show()
# %%

# %%



import itertools
import os
import typing
from enum import Enum
from typing import Any, Callable, Type

import dask
import dask.array as da
import napari
import numpy as np
import pydantic
import tifffile
from PyQt5 import QtCore
from PyQt5.QtWidgets import QWidget
from qtpy import QtGui, QtWidgets
from scipy import ndimage
from scipy import signal    
from slugify import slugify
from superqt import (
    QDoubleRangeSlider,
    QEnumComboBox,
    QLabeledDoubleRangeSlider,
    QLabeledDoubleSlider,
)
from superqt.utils import thread_worker


from faser.env import get_asset_file
from faser.generators.base import AberrationFloat, PSFConfig
from faser.generators.vectorial.stephane.tilted_coverslip import generate_psf
from faser.napari.widgets.fields import generate_single_widgets_from_model


class HelperTab(QtWidgets.QWidget):
    def __init__(
        self,
        viewer: napari.Viewer,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.viewer = viewer
        self.mylayout = QtWidgets.QVBoxLayout()
        self.mylayout.setContentsMargins(0, 0, 0, 0)
        self.mylayout.setSpacing(1)
        self.setLayout(self.mylayout)


# Step 1: Create a worker class
class ExportWorker(QtCore.QObject):
    finished = QtCore.pyqtSignal()
    progress = QtCore.pyqtSignal(int)

    def __init__(self, layers, export_dir):
        super().__init__()
        self.layers = layers
        self.export_dir = export_dir

    def export_layer_with_config_data_to_file(
        self, data, export_dir, layer_name, config
    ):
        export_file_dir = os.path.join(export_dir, slugify(layer_name))
        # export_file_dir = os.path.join(export_dir, "test")
        os.makedirs(export_file_dir, exist_ok=True)
        with open(
            os.path.join(export_file_dir, "config.txt"), "w", encoding="utf-8"
        ) as f:
            f.write(config.json())

        tifffile.imsave(os.path.join(export_file_dir, "psf.tif"), data)
        print("Exported")

    def run(self):
        """Long-running task."""
        print("Running")
        for layer in self.layers:
            if layer.metadata.get("is_psf", False) is True:
                if layer.metadata.get("is_batch", False) is True:
                    first_dim = layer.data.shape[0]
                    for i in range(first_dim):

                        self.progress.emit(i + 1)
                        self.export_layer_with_config_data_to_file(
                            layer.data[i, :, :, :],
                            self.export_dir,
                            layer.name,
                            layer.metadata["configs"][i],
                        )
                else:
                    print("Exporting this one")
                    self.export_layer_with_config_data_to_file(
                        layer.data,
                        self.export_dir,
                        layer.name,
                        layer.metadata["config"],
                    )

        self.finished.emit()


@thread_worker
def export_layer_with_config_data_to_file(data, export_dir, layer_name, config):
    export_file_dir = os.path.join(export_dir, "test")
    os.makedirs(export_file_dir, exist_ok=True)
    with open(os.path.join(export_file_dir, "config.txt"), "w", encoding="utf-8") as f:
        f.write(config.json())

    tifffile.imsave(os.path.join(export_file_dir, "psf.tif"), data)
    print("Exported")


class ExportTab(HelperTab):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.show = QtWidgets.QPushButton("Select PSF Layers")
        self.show.setEnabled(False)
        self.show.clicked.connect(self.export_pressed)
        self.mylayout.addStretch()

        self.viewer.layers.selection.events.connect(self.update_selection)

        self.mylayout.addWidget(self.show)

    def on_worker_done(self):
        print("done")
        self.show.setEnabled(True)

    def on_worker_progress(self, value):
        print(value)

    def update_selection(self, event):
        selection = self.viewer.layers.selection

        if not selection:
            self.show.setEnabled(False)
            self.show.setText("Select PSF")

        else:
            layers = [
                layer for layer in selection if layer.metadata.get("is_psf", False)
            ]
            if len(layers) == 0:
                self.show.setEnabled(False)
                self.show.setText("Select PSF Layers")

            else:
                self.show.setEnabled(True)
                self.show.setText("Export PSF" if len(layers) == 1 else "Export PSFs")

        print(self.viewer.layers.selection.active)

    def export_layer_with_config_data_to_file(data, export_dir, layer_name, config):
        export_file_dir = os.path.join(export_dir, layer_name)
        os.makedirs(export_file_dir, exist_ok=True)
        with open(
            os.path.join(export_file_dir, "config.txt"), "w", encoding="utf-8"
        ) as f:
            f.write(config.json())

        tifffile.imsave(os.path.join(export_file_dir, "psf.tif"), data)
        print("Exported")

    def export_active_selection(self, export_dir):
        layers = []
        for layer in self.viewer.layers.selection:
            if layer.metadata.get("is_psf", False) == True:
                layers.append(layer)

        self.thread = QtCore.QThread()
        self.worker = ExportWorker(layers, export_dir)
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)
        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)
        self.worker.progress.connect(self.on_worker_progress)
        self.thread.start()

    def export_pressed(self):
        export_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Select Directory"
        )

        if export_dir:
            self.export_active_selection(export_dir=export_dir)


class SpaceModel(pydantic.BaseModel):
    x_size: int = 1000
    y_size: int = 1000
    z_size: int = 10
    dots: int = 50


class SampleTab(HelperTab):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.show = QtWidgets.QPushButton("Create space")
        self.show.clicked.connect(self.generate_space)

        self.managed_widgets = generate_single_widgets_from_model(
            SpaceModel,
            callback=self.callback,
            range_callback=None,
            parent=self,
        )

        print(self.managed_widgets)

        for widget in self.managed_widgets:
            widget.init_ui_helper()
            self.mylayout.addWidget(widget)

        self.mylayout.addStretch()

        self.mylayout.addWidget(self.show)
        self.space_model = SpaceModel()

    def callback(self, name, value):
        split = name.split(".")
        if len(split) > 1:
            self.space_model.__getattribute__(split[0]).__setattr__(split[1], value)
        else:
            self.space_model.__setattr__(name, value)

    def show_wavefront(self):
        raise NotImplementedError()

    def generate_space(self):
        x = np.random.randint(0, self.space_model.x_size, size=(self.space_model.dots))
        y = np.random.randint(0, self.space_model.y_size, size=(self.space_model.dots))
        z = np.random.randint(0, self.space_model.z_size, size=(self.space_model.dots))

        M = np.zeros(
            (
                self.space_model.z_size,
                self.space_model.y_size,
                self.space_model.y_size,
            )
        )
        for p in zip(z, x, y):
            M[p] = 1

        self.viewer.add_image(M, name="Space")


class EffectiveModel(pydantic.BaseModel):
    Isat: float = pydantic.Field(default=0.1, lt=1, gt=0)


class EffectiveTab(HelperTab):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.managed_widgets = generate_single_widgets_from_model(
            EffectiveModel,
            callback=self.callback,
            range_callback=None,
            parent=self,
        )

        print(self.managed_widgets)

        for widget in self.managed_widgets:
            widget.init_ui_helper_eff()
            self.mylayout.addWidget(widget)

        self.label = QtWidgets.QLabel("Make Effective PSF")
        self.show = QtWidgets.QPushButton("Select exactly 2 PSFs")
        self.show.setEnabled(False)
        self.show.clicked.connect(self.make_effective_psf)

        self.show_alternate = QtWidgets.QPushButton("Select exactly 2 PSFs")
        self.show_alternate.setEnabled(False)
        self.show_alternate.clicked.connect(self.make_effective_psf_alternate)

        hlayout = QtWidgets.QHBoxLayout()
        hlayout.addWidget(self.show)
        hlayout.addWidget(self.show_alternate)

        # self.mylayout.addStretch()
        self.mylayout.addWidget(self.label)
        self.mylayout.addStretch()
        self.mylayout.addLayout(hlayout)

        self.effective_model = EffectiveModel()

        self.viewer.layers.selection.events.connect(self.update_selection)

    def callback(self, name, value):
        split = name.split(".")
        if len(split) > 1:
            self.effective_model.__getattribute__(split[0]).__setattr__(split[1], value)
        else:
            self.effective_model.__setattr__(name, value)

    def make_effective_psf(self):
        I_sat = self.effective_model.Isat
        psf_layers = list(
            layer
            for layer in self.viewer.layers.selection
            if layer.metadata.get("is_psf", True)
        )

        assert len(psf_layers) == 2, "Select exactly 2 PSFs"

        psf_layer_one = psf_layers[0]  # Excitation PSF
        psf_layer_two = psf_layers[1]  # Depletion PSF
        new_psf = np.multiply(psf_layer_one.data, np.exp(-psf_layer_two.data / I_sat))

        return self.viewer.add_image(
            new_psf,
            name=f"Combined PSF {psf_layer_one.name} {psf_layer_two.name}",
            metadata={"is_psf": True},
            colormap="viridis",
        )

    def make_effective_psf_alternate(self):
        I_sat = self.effective_model.Isat
        psf_layers = list(
            layer
            for layer in self.viewer.layers.selection
            if layer.metadata.get("is_psf", True)
        )

        assert len(psf_layers) == 2, "Select exactly 2 PSFs"

        psf_layer_one = psf_layers[1]  # Excitation PSF
        psf_layer_two = psf_layers[0]  # Depletion PSF
        new_psf = np.multiply(psf_layer_one.data, np.exp(-psf_layer_two.data / I_sat))

        return self.viewer.add_image(
            new_psf,
            name=f"Combined PSF {psf_layer_one.name} {psf_layer_two.name}",
            metadata={"is_psf": True},
            colormap="viridis",
        )

    def update_selection(self, event):
        selection = self.viewer.layers.selection

        if not selection:
            self.show.setEnabled(False)
            self.show.setText("Select 2 PSFs")
            self.show_alternate.setEnabled(False)
            self.show_alternate.setText("Select 2 PSFs")

        psf_layers = list(
            layer
            for layer in self.viewer.layers.selection
            if layer.metadata.get("is_psf", True)
        )

        if len(psf_layers) != 2:
            self.show.setEnabled(False)
            self.show.setText("Select 2 PSFs")
            self.show_alternate.setEnabled(False)
            self.show_alternate.setText("Select 2 PSFs")
            return

        layer_one = psf_layers[0]
        layer_two = psf_layers[1]

        self.show.setText(f"{layer_one.name} -> {layer_two.name}")
        self.show_alternate.setText(f"{layer_two.name} -> {layer_one.name}")
        self.show.setEnabled(True)
        self.show_alternate.setEnabled(True)


class ConvolveModel(pydantic.BaseModel):
    pass


# Step 1: Create a worker class
class ConvolveWorker(QtCore.QObject):
    finished = QtCore.pyqtSignal(object)
    progress = QtCore.pyqtSignal(int)

    def __init__(self, image_data, psf_data):
        super().__init__()
        self.image_data = image_data
        self.psf_data = psf_data

    def export_layer_with_config_data_to_file(
        self, data, export_dir, layer_name, config
    ):
        export_file_dir = os.path.join(export_dir, slugify(layer_name))
        os.makedirs(export_file_dir, exist_ok=True)
        with open(
            os.path.join(export_file_dir, "config.txt"), "w", encoding="utf-8"
        ) as f:
            f.write(config.json())

        tifffile.imsave(os.path.join(export_file_dir, "psf.tif"), data)
        print("Exported")

    def run(self):
        """Long-running task."""
        if self.image_data.ndim == 2:
            psf_data = self.psf_data[self.psf_data.shape[0] // 2, :, :]

            con = signal.convolve(
                self.image_data, psf_data, mode="constant", cval=0.0, origin=0
            )

            self.finished.emit(con)
            return

        con = signal.convolve(
            self.image_data, self.psf_data, mode="full", method="fft"
        )

        self.finished.emit(con)


class ConvolveTab(HelperTab):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.managed_widgets = generate_single_widgets_from_model(
            ConvolveModel,
            callback=self.callback,
            range_callback=None,
            parent=self,
        )

        print(self.managed_widgets)

        for widget in self.managed_widgets:
            widget.init_ui_helper()
            self.mylayout.addWidget(widget)

        self.show = QtWidgets.QPushButton("Select image and PSF")
        self.show.setEnabled(False)
        self.show.clicked.connect(self.convolve_psf)
        self.mylayout.addStretch()

        self.mylayout.addWidget(self.show)

        self.effective_model = EffectiveModel()

        self.viewer.layers.selection.events.connect(self.update_selection)

    def callback(self, name, value):
        split = name.split(".")
        if len(split) > 1:
            self.effective_model.__getattribute__(split[0]).__setattr__(split[1], value)
        else:
            self.effective_model.__setattr__(name, value)

    def on_worker_done(self, con: np.array):
        self.show.setText("Select image and PSF")
        return self.viewer.add_image(
            con.squeeze(),
            name=f"Convoled Image",
        )

    def on_worker_progress(self, value):
        print(value)

    def convolve_psf(self):
        print("Convolve PSF and image")
        psf_layer = next(
            layer
            for layer in self.viewer.layers.selection
            if layer.metadata.get("is_psf", False)
        )
        image_layer = next(
            layer
            for layer in self.viewer.layers.selection
            if not layer.metadata.get("is_psf", False)
        )

        image_data = image_layer.data
        psf_data = psf_layer.data

        self.show.setText("Convolviing...")

        self.thread = QtCore.QThread()
        self.worker = ConvolveWorker(image_data, psf_data)
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)
        self.worker.finished.connect(self.on_worker_done)
        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)
        self.worker.progress.connect(self.on_worker_progress)
        self.thread.start()

    def update_selection(self, event):
        selection = self.viewer.layers.selection

        if not selection:
            self.show.setEnabled(False)
            self.show.setText("Select a PSF and the Image")

        else:
            layers = [
                layer for layer in selection if layer.metadata.get("is_psf", False)
            ]
            if len(layers) != 1:
                self.show.setEnabled(False)
                self.show.setText("Select only one PSF ")

            else:
                self.show.setEnabled(True)
                self.show.setText("Convolve Image")


class InspectTab(HelperTab):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.show = QtWidgets.QPushButton("Show Intensity")
        self.show.clicked.connect(self.show_wavefront)

        self.showp = QtWidgets.QPushButton("Show Phasemask")
        self.showp.clicked.connect(self.show_wavefront)

        self.viewer.layers.selection.events.connect(self.update_selection)

        self.mylayout.addWidget(self.show)
        self.mylayout.addWidget(self.showp)

    def update_selection(self, event):
        print(self.viewer.layers.selection.active)

    def show_wavefront(self):
        raise NotImplementedError()


class HelperWidget(QtWidgets.QWidget):
    def __init__(self, viewer: napari.Viewer, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.viewer = viewer
        self.effective_tab = EffectiveTab(
            self.viewer,
        )
        self.sample_tab = SampleTab(
            self.viewer,
        )
        self.inspect_tab = InspectTab(
            self.viewer,
        )
        self.convolve_tab = ConvolveTab(
            self.viewer,
        )
        self.export_tab = ExportTab(
            self.viewer,
        )

        layout = QtWidgets.QGridLayout()
        tabwidget = QtWidgets.QTabWidget()
        tabwidget.addTab(self.effective_tab, "Effective")
        tabwidget.addTab(self.sample_tab, "Sample")
        # tabwidget.addTab(self.inspect_tab, "Inspect")
        tabwidget.addTab(self.convolve_tab, "Convolve")
        tabwidget.addTab(self.export_tab, "Export")
        layout.addWidget(tabwidget, 0, 0)

        self.setLayout(layout)



#%%
from faser.generators.base import PSFGeneratorConfig
from faser.generators.scalar.phasenet import PhaseNetPSFGenerator
from faser.generators.vectorial.stephane import StephanePSFGenerator
from faser.generators.scalar.gibson_lanny import GibsonLannyPSFGenerator
import matplotlib.pyplot as plt
import numpy as np
import matplotlib
import napari

config = PSFGeneratorConfig(Nx=65, Ny=65, Nz=65)


from faser.io import save


save(PhaseNetPSFGenerator, config, "test_phase_net.zarr")
save(StephanePSFGenerator, config, "test_stephane.zarr")



# import contextlib
# from faser.napari.gui import generate_psf_gui
from faser.napari.main import main

# import napari
# import numpy as np
import argparse
import os

if __name__ == "__main__":

    os.environ["NAPARI_ASYNC"] = "1"
    parser = argparse.ArgumentParser()
    args = parser.parse_args()

    main()



# import typing
import itertools
import os
import time

# from typing import Callable, Type, Any
# import typing
# from pydantic.fields import ModelField
# from superqt import QEnumComboBox
from enum import Enum

import dask
import dask.array as da
import napari

# from pydantic.types import ConstrainedFloat
import numpy as np
from PyQt5 import QtCore

# from PyQt5.QtWidgets import QWidget
from qtpy import QtGui, QtWidgets
from qtpy.QtWidgets import QStyle

# from superqt import QDoubleRangeSlider, QLabeledDoubleRangeSlider, QLabeledDoubleSlider
# import pydantic
from faser.env import get_asset_file
from faser.generators.base import AberrationFloat, PSFConfig
from faser.generators.utils import polar_phase_mask, polar_to_cartesian
from faser.generators.vectorial.stephane.tilted_coverslip import (
    generate_aberration,
    generate_intensity_profile,
    generate_phase_mask,
    generate_psf,
)

from .fields import build_key_filter, generate_single_widgets_from_model
from .mpl_canvas import BeamDialog, MatplotlibDialog, PhaseMaskDialog, WavefrontDialog


class ScrollableWidget(QtWidgets.QWidget):
    def __init__(
        self,
        viewer: napari.Viewer,
        *args,
        filter_fields=None,
        callback=None,
        range_callback=None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.viewer = viewer
        self.managed_widgets = generate_single_widgets_from_model(
            PSFConfig,
            callback=callback,
            range_callback=range_callback,
            parent=self,
            filter_fields=filter_fields,
        )

        self.mylayout = QtWidgets.QVBoxLayout()
        self.mylayout.setContentsMargins(0, 0, 0, 0)
        self.mylayout.setSpacing(1)

        for widget in self.managed_widgets:
            widget.init_ui()
            self.mylayout.addWidget(widget)

        self.active_base_model = PSFConfig()
        self.setLayout(self.mylayout)

    def model_value_changed(self, name, value):
        self.active_base_model.__setattr__(name, value)
        print(self.active_base_model.model_dump_json())

    def reset(self):
        for widget in self.managed_widgets:
            widget.reset_default()


class SampleTab(QtWidgets.QWidget):
    def __init__(
        self,
        viewer: napari.Viewer,
        main=None,
        *args,
        filter_fields=None,
        callback=None,
        range_callback=None,
        image="placeholder.png",
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)


        self.viewer = viewer
        self.main = main

        self.scroll = (
            QtWidgets.QScrollArea()
        )  # Scroll Area which contains the widgets, set as the centralWidget
        self.widget = ScrollableWidget(
            self.viewer,
            callback=callback,
            range_callback=range_callback,
            filter_fields=filter_fields,
        )  # Widget that contains the collection of Vertical Box

        # Scroll Area Properties
        self.scroll.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        self.scroll.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scroll.setWidgetResizable(True)
        self.scroll.setWidget(self.widget)

        self.mylayout = QtWidgets.QVBoxLayout()
        self.mylayout.setContentsMargins(10,10, 10, 10)
        self.mylayout.setSpacing(1)

        if image is not None:
            self.im = QtGui.QPixmap(get_asset_file(image))
            self.im = self.im.scaledToHeight(300)
            self.label = QtWidgets.QLabel()
            self.label.setPixmap(self.im)
            self.mylayout.addWidget(self.label)

        self.mylayout.addWidget(self.scroll)

        self.setLayout(self.mylayout)

    def reset(self):
        self.widget.reset()


@dask.delayed
def lazy_generate_psf(config: PSFConfig):
    return generate_psf(config)


def generate_wavefront(config: PSFConfig):
    num_radii = 150
    num_angles = 300
    polar_mask = polar_phase_mask(num_radii, num_angles)

    # Convert to Cartesian coordinates
    cartesian_mask = polar_to_cartesian(polar_mask)
    return cartesian_mask


class AbberationTab(SampleTab):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.show_button = QtWidgets.QPushButton("Show Wavefront")
        self.show_button.clicked.connect(self.show_wavefront)
        self.wavefront_dialog = WavefrontDialog("Wavefront", parent=self)

        self.mylayout.addWidget(self.show_button)

    def show_wavefront(self):
        self.wavefront_dialog.update(
            generate_aberration(self.main.active_base_model), "Wavefront"
        )
        self.wavefront_dialog.show()


class BeamTab(SampleTab):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.show_button = QtWidgets.QPushButton("Show Intensity")
        self.show_button.clicked.connect(self.show_intensity)

        self.showp_button = QtWidgets.QPushButton("Show Phase Mask")
        self.showp_button.clicked.connect(self.show_phase_mask)

        self.intensity_dialog = BeamDialog("Intensity", parent=self)
        self.phase_mask_dialog = PhaseMaskDialog("Phase Mask", parent=self)

        self.mylayout.addWidget(self.show_button)
        self.mylayout.addWidget(self.showp_button)

    def show_intensity(self):
        self.intensity_dialog.update(
            generate_intensity_profile(self.main.active_base_model), "Intensity"
        )
        self.intensity_dialog.show()

    def show_phase_mask(self):
        self.phase_mask_dialog.update(
            generate_phase_mask(self.main.active_base_model), "Phase mask"
        )
        self.phase_mask_dialog.show()


simulation_set = [
    "L_obs_XY",
    "L_obs_Z",
    "Ntheta",
    "Nphi",
    "Nxy",
    "Nz",
    "Normalize",  # ,
    # "add_detector_poisson_noise"
]

geometry_set = [
    "NA",
    "WD",
    "n1",
    "n2",
    "n3",
    "Thickness",
    "Collar",
    "Depth",
    "Tilt",
    "Window",
    "Wind_Radius",
    "Wind_Depth",
]

beam_set = [
    "Mode",
    "Polarization",
    "Wavelength",
    "Waist",
    "Ampl_offset_x",
    "Ampl_offset_y",
    "Psi",
    "Epsilon",
    "Ring_Radius",
    "VC",
    "RC",
    "Mask_offset_x",
    "Mask_offset_y",
    "p",
]

aberration_set = [
    "a0",
    "a1",
    "a2",
    "a3",
    "a4",
    "a5",
    "a6",
    "a7",
    "a8",
    "a9",
    "a12",
    "Aberration_offset_x",
    "Aberration_offset_y",
]


class MainWidget(QtWidgets.QWidget):
    def __init__(self, viewer: napari.Viewer, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.viewer = viewer
        self.simulation_tab = SampleTab(
            self.viewer,
            main=self,
            image="Figure_simu.png",
            callback=self.callback,
            range_callback=self.range_callback,
            filter_fields=build_key_filter(simulation_set),
        )
        self.geometry_tab = SampleTab(
            self.viewer,
            main=self,
            image="Figure_geom.png",
            callback=self.callback,
            range_callback=self.range_callback,
            filter_fields=build_key_filter(geometry_set),
        )
        self.beam_tab = BeamTab(
            self.viewer,
            main=self,
            image="Figure_beam.png",
            callback=self.callback,
            range_callback=self.range_callback,
            filter_fields=build_key_filter(beam_set),
        )
        self.aberration_tab = AbberationTab(
            self.viewer,
            main=self,
            image="Figure_ab.png",
            callback=self.callback,
            range_callback=self.range_callback,
            filter_fields=build_key_filter(aberration_set),
        )

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(10, 10, 10, 10)
        layout.setSpacing(1)
        self.setLayout(layout)
        tabwidget = QtWidgets.QTabWidget()
        tabwidget.addTab(self.simulation_tab, "Simulation")
        tabwidget.addTab(self.geometry_tab, "Geometry")
        tabwidget.addTab(self.beam_tab, "Beam")
        tabwidget.addTab(self.aberration_tab, "Abberation")

        self.generate = QtWidgets.QPushButton("Generate")

        self.generate.clicked.connect(self.generate_psf)
        self.generate.setMinimumHeight(20)
        self.generate.setMaximumHeight(20)


        load_pixmal = QtGui.QPixmap(get_asset_file("save.png")).scaled(15,15, QtCore.Qt.KeepAspectRatio, QtCore.Qt.SmoothTransformation)

        load_icon = QtGui.QIcon(load_pixmal)
        self.loadb = QtWidgets.QPushButton()
        self.loadb.setIcon(load_icon)
        self.loadb.clicked.connect(self.load_model)
        self.loadb.setMinimumWidth(20)
        self.loadb.setMaximumWidth(20)
        self.loadb.setMinimumHeight(20)
        self.loadb.setMaximumHeight(20)

        save_icon = QtGui.QIcon(QtGui.QPixmap(get_asset_file("load.png")).scaled(15,15, QtCore.Qt.KeepAspectRatio, QtCore.Qt.SmoothTransformation))
        self.saveb = QtWidgets.QPushButton()
        self.saveb.setIcon(save_icon)
        self.saveb.clicked.connect(self.save_model)
        self.saveb.setMinimumWidth(20)
        self.saveb.setMaximumWidth(20)
        self.saveb.setMinimumHeight(20)
        self.saveb.setMaximumHeight(20)

        refresh_icon = QtGui.QIcon(QtGui.QPixmap(get_asset_file("refresh.png")).scaled(15,15, QtCore.Qt.KeepAspectRatio, QtCore.Qt.SmoothTransformation))
        self.resetb = QtWidgets.QPushButton()
        self.resetb.setIcon(refresh_icon)
        self.resetb.clicked.connect(self.reset_values)
        self.resetb.setMinimumWidth(20)
        self.resetb.setMaximumWidth(20)
        self.resetb.setMinimumHeight(20)
        self.resetb.setMaximumHeight(20)

        hlayout = QtWidgets.QHBoxLayout()
        hlayout.setContentsMargins(10, 10, 10, 10)
        hlayout.addWidget(self.generate)
        hlayout.addWidget(self.loadb)
        hlayout.addWidget(self.saveb)
        hlayout.addWidget(self.resetb)

        layout.addWidget(tabwidget)

        layout.addLayout(hlayout)
        self.active_base_model = PSFConfig()

        self.active_batchers = {}

    def reset_values(self):
        self.simulation_tab.reset()
        self.geometry_tab.reset()
        self.beam_tab.reset()
        self.aberration_tab.reset()
        self.active_batchers = {}

    def callback(self, name, value):
        split = name.split(".")
        if len(split) > 1:
            self.active_base_model.__getattribute__(split[0]).__setattr__(
                split[1], value
            )
        else:
            self.active_base_model.__setattr__(name, value)

    def range_callback(self, name, value):
        if value == None:
            if name in self.active_batchers:
                del self.active_batchers[name]

        else:
            self.active_batchers[name] = value

        print(self.active_batchers)

        self.update_ui()

    def calculate_batcher_length(self):
        length = 1
        for key, value in self.active_batchers.items():
            length *= len(value.to_list())

        return length

    def update_ui(self):
        if len(self.active_batchers) > 0:
            self.generate.setText(f"Generate {self.calculate_batcher_length()} PSFs")

        else:
            self.generate.setText("Generate")
            print(self.active_batchers)

    def reinit_ui(self):
        for key, value in self.active_batchers.items():
            self.update_default(key, value)

    def save_model(self):
        default = os.path.join(os.getcwd(), "psf_config.json")
        name, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save File", default, "Config Files (*.json)"
        )
        if name:
            file = open(name, "w")
            text = self.active_base_model.json()
            file.write(text)
            file.close()

    def load_model(self):
        name, _ = QtWidgets.QFileDialog.getOpenFileName(
            self, "Open File", "", "Config Files (*.json)"
        )
        if name:
            file = open(name, "r")
            text = file.read()
            self.active_base_model = PSFConfig.parse_raw(text)
            self.reinit_ui()

    def start_map(self):
        mapper = self.active_batchers

        permutations = list(
            itertools.product(*map(lambda x: x.to_list(), mapper.values()))
        )

        models = []

        for i in permutations:
            values = {k: v for k, v in zip(mapper.keys(), i)}
            new_model = self.active_base_model.copy(update=values)
            models.append(new_model)
            print(new_model)

        t = [
            da.from_delayed(
                lazy_generate_psf(new_model),
                (new_model.Nz, new_model.Nxy, new_model.Nxy),
                dtype=np.float64,
            )
            for new_model in models
        ]
        t = da.stack(t, axis=0)

        reshape_start = [
            len(value.to_list()) for key, value in self.active_batchers.items()
        ]

        if len(reshape_start) < 3:
            t = t.reshape(
                reshape_start
                + [
                    self.active_base_model.Nz,
                    self.active_base_model.Nxy,
                    self.active_base_model.Nxy,
                ],
                merge_chunks=False,
            )

        return self.viewer.add_image(
            t,
            name="Batch PSF",
            metadata={"is_psf": True, "configs": models, "is_batch": True},
            multiscale=False,
            colormap="viridis",
        )

    def generate_psf(self):

        start = time.time()

        self.generate.setText("Generating...")

        if len(self.active_batchers) > 0:
            self.start_map()

        else:

            config = self.active_base_model
            psf = generate_psf(config)
            print(psf.max())
            self.viewer.add_image(
                psf,
                name=f"PSF ",
                metadata={"is_psf": True, "config": config, "is_batch": False},
                colormap="viridis",
            )
            self.generate.setText("Generate")

        end = time.time()
        print(end - start)



import sys
import matplotlib

matplotlib.use("Qt5Agg")

from PyQt5 import QtCore, QtWidgets

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
import napari


class MplCanvas(FigureCanvasQTAgg):
    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = fig.add_subplot(111)
        super(MplCanvas, self).__init__(fig)


class TopBar(QtWidgets.QWidget):
    def __init__(self, viewer: napari.Viewer, *args, **kwargs):
        super(TopBar, self).__init__(*args, **kwargs)

        # Create the maptlotlib FigureCanvas object,
        # which defines a single set of axes as self.axes.
        self.layout = QtWidgets.QHBoxLayout()
        sc = MplCanvas(self, width=5, height=4, dpi=100)
        sc.axes.plot([0, 1, 2, 3, 4], [10, 1, 20, 3, 40])

        self.layout.addWidget(sc)
        self.setLayout(self.layout)
        self.show()



import numpy as np


def ger_sax(img, max_iter):
    h, w = img.shape
    pm_s = np.random.rand(h, w)
    pm_f = np.ones((h, w))
    am_s = np.sqrt(img)
    am_f = np.ones((h, w))

    signal_s = am_s * np.exp(pm_s * 1j)

    for iter in range(max_iter):
        signal_f = np.fft.fft2(signal_s)
        pm_f = np.angle(signal_f)
        signal_f = am_f * np.exp(pm_f * 1j)
        signal_s = np.fft.ifft2(signal_f)
        pm_s = np.angle(signal_s)
        signal_s = am_s * np.exp(pm_s * 1j)

    pm = pm_f
    return pm



import os
import tifffile
from faser.generators.base import PSFGeneratorConfig

dir_path = os.path.dirname(os.path.realpath(__file__))
generator_path = os.path.join(dir_path, "PSFGenerator.jar")
config_path = "config.txt"


def generate_psf(config: PSFGeneratorConfig):

    with open(config_path, "w") as f:
        f.write(
            f"""
#PSFGenerator
#Sun Jan 14 21:27:57 CET 2018
PSF-shortname=GL
ResLateral=100.0
ResAxial=250.0
NY={config.Ny}
NX={config.Nx}
NZ={config.Nz}
Type=32-bits
NA={config.numerical_aperature}
LUT=Grays
Lambda=610.0
Scale=Linear
psf-BW-NI=1.5
psf-BW-accuracy=Good
psf-RW-NI=1.5
psf-RW-accuracy=Good
psf-GL-NI={config.refractive_index}
psf-GL-NS=1.33
psf-GL-accuracy=Good
psf-GL-ZPos=2000.0
psf-GL-TI={config.working_distance * 10e5}
psf-TV-NI=1.5
psf-TV-ZPos=2000.0
psf-TV-TI=150.0
psf-TV-NS=1.0
psf-Circular-Pupil-defocus=100.0
psf-Circular-Pupil-axial=Linear
psf-Circular-Pupil-focus=0.0
psf-Oriented-Gaussian-axial=Linear
psf-Oriented-Gaussian-focus=0.0
psf-Oriented-Gaussian-defocus=100.0
psf-Astigmatism-focus=0.0
psf-Astigmatism-axial=Linear
psf-Astigmatism-defocus=100.0
psf-Defocus-DBot=30.0
psf-Defocus-ZI=2000.0
psf-Defocus-DTop=30.0
psf-Defocus-DMid=1.0
psf-Defocus-K=275.0
psf-Cardinale-Sine-axial=Linear
psf-Cardinale-Sine-defocus=100.0
psf-Cardinale-Sine-focus=0.0
psf-Lorentz-axial=Linear
psf-Lorentz-focus=0.0
psf-Lorentz-defocus=100.0
psf-Koehler-dMid=3.0
psf-Koehler-dTop=1.5
psf-Koehler-n1=1.0
psf-Koehler-n0=1.5
psf-Koehler-dBot=6.0
psf-Double-Helix-defocus=100.0
psf-Double-Helix-axial=Linear
psf-Double-Helix-focus=0.0
psf-Gaussian-axial=Linear
psf-Gaussian-focus=0.0
psf-Gaussian-defocus=100.0
psf-Cosine-axial=Linear
psf-Cosine-focus=0.0
psf-Cosine-defocus=100.0
psf-VRIGL-NI=1.5
psf-VRIGL-accuracy=Good
psf-VRIGL-NS2=1.4
psf-VRIGL-NS1=1.33
psf-VRIGL-TG=170.0
psf-VRIGL-NG=1.5
psf-VRIGL-TI=150.0
psf-VRIGL-RIvary=Linear
psf-VRIGL-ZPos=2000.0
"""
        )

    print(f"java -cp {generator_path} {config_path}")
    os.system(f"java -cp {generator_path} PSFGenerator {config_path}")

    x = tifffile.imread("PSF GL.tif")
    os.remove("PSF GL.tif")
    os.remove(config_path)
    return x



import numpy as np
from functools import lru_cache
from scipy.special import binom


def nm_to_noll(n, m):
    j = (n * (n + 1)) // 2 + abs(m)
    if m > 0 and n % 4 in (0, 1):
        return j
    if m < 0 and n % 4 in (2, 3):
        return j
    if m >= 0 and n % 4 in (2, 3):
        return j + 1
    if m <= 0 and n % 4 in (0, 1):
        return j + 1
    assert False


def nm_to_ansi(n, m):
    return (n * (n + 2) + m) // 2


def nm_normalization(n, m):
    """the norm of the zernike mode n,m in born/wolf convetion
    i.e. sqrt( \int | z_nm |^2 )
    """
    return np.sqrt((1.0 + (m == 0)) / (2.0 * n + 2))


def nm_polynomial(n, m, rho, theta, normed=True):
    """returns the zernike polyonimal by classical n,m enumeration
    if normed=True, then they form an orthonormal system
        \int z_nm z_n'm' = delta_nn' delta_mm'
        and the first modes are
        z_nm(0,0)  = 1/sqrt(pi)*
        z_nm(1,-1) = 1/sqrt(pi)* 2r cos(phi)
        z_nm(1,1)  = 1/sqrt(pi)* 2r sin(phi)
        z_nm(2,0)  = 1/sqrt(pi)* sqrt(3)(2 r^2 - 1)
        ...
        z_nm(4,0)  = 1/sqrt(pi)* sqrt(5)(6 r^4 - 6 r^2 +1)
        ...
    if normed =False, then they follow the Born/Wolf convention
        (i.e. min/max is always -1/1)
        \int z_nm z_n'm' = (1.+(m==0))/(2*n+2) delta_nn' delta_mm'
        z_nm(0,0)  = 1
        z_nm(1,-1) = r cos(phi)
        z_nm(1,1)  =  r sin(phi)
        z_nm(2,0)  = (2 r^2 - 1)
        ...
        z_nm(4,0)  = (6 r^4 - 6 r^2 +1)
    """
    if abs(m) > n:
        raise ValueError(" |m| <= n ! ( %s <= %s)" % (m, n))

    if (n - m) % 2 == 1:
        return 0 * rho + 0 * theta

    radial = 0
    m0 = abs(m)

    for k in range((n - m0) // 2 + 1):
        radial += (
            (-1.0) ** k
            * binom(n - k, k)
            * binom(n - 2 * k, (n - m0) // 2 - k)
            * rho ** (n - 2 * k)
        )

    radial *= rho <= 1.0

    if normed:
        prefac = 1.0 / nm_normalization(n, m)
    else:
        prefac = 1.0
    if m >= 0:
        return prefac * radial * np.cos(m0 * theta)
    else:
        return prefac * radial * np.sin(m0 * theta)


@lru_cache(maxsize=32)
def rho_theta(size):
    r = np.linspace(-1, 1, size)
    X, Y = np.meshgrid(r, r, indexing="ij")
    rho = np.hypot(X, Y)
    theta = np.arctan2(Y, X)
    return rho, theta


@lru_cache(maxsize=32)
def outside_mask(size):
    rho, theta = rho_theta(size)
    return nm_polynomial(0, 0, rho, theta, normed=False) < 1


def dict_to_list(kv):
    max_key = max(kv.keys())
    out = [0] * (max_key + 1)
    for k, v in kv.items():
        out[k] = v
    return out


def ensure_dict(values, order):
    if isinstance(values, dict):
        return values
    if isinstance(values, np.ndarray):
        values = tuple(values.ravel())
    if isinstance(values, (tuple, list)):
        order = str(order).lower()
        assert order in (
            "noll",
            "ansi",
        ), "Could not identify the Zernike nomenclature/order"
        offset = 1 if order == "noll" else 0
        indices = range(offset, offset + len(values))
        return dict(zip(indices, values))
    raise ValueError("Could not identify the data type for dictionary formation")


class Zernike:

    """
    Encapsulates Zernike polynomials
    :param index: string, integer or tuple, index of Zernike polynomial e.g. 'defocus', 4, (2,2)
    :param oder: string, defines the Zernike nomenclature if index is an integer, eg noll or ansi, default is noll
    """

    _ansi_names = [
        "piston",
        "tilt",
        "tip",
        "oblique astigmatism",
        "defocus",
        "vertical astigmatism",
        "vertical trefoil",
        "vertical coma",
        "horizontal coma",
        "oblique trefoil",
        "oblique quadrafoil",
        "oblique secondary astigmatism",
        "primary spherical",
        "vertical secondary astigmatism",
        "vertical quadrafoil",
    ]
    _nm_pairs = set((n, m) for n in range(200) for m in range(-n, n + 1, 2))
    _noll_to_nm = dict(zip((nm_to_noll(*nm) for nm in _nm_pairs), _nm_pairs))
    _ansi_to_nm = dict(zip((nm_to_ansi(*nm) for nm in _nm_pairs), _nm_pairs))

    def __init__(self, index, order="noll"):
        super().__setattr__("_mutable", True)
        if isinstance(index, str):
            if index.isdigit():
                index = int(index)
            else:
                name = index.lower()
                assert (
                    name in self._ansi_names
                ), "Your input for index is string : Could not identify the name of Zernike polynomial"
                index = self._ansi_names.index(name)
                order = "ansi"

        if isinstance(index, (list, tuple)) and len(index) == 2:
            self.n, self.m = int(index[0]), int(index[1])
            assert (
                self.n,
                self.m,
            ) in self._nm_pairs, "Your input for index is list/tuple : Could not identify the n,m order of Zernike polynomial"
        elif isinstance(index, int):
            order = str(order).lower()
            assert order in (
                "noll",
                "ansi",
            ), "Your input for index is int : Could not identify the Zernike nomenclature/order"

            if order == "noll":
                assert (
                    index in self._noll_to_nm
                ), "Your input for index is int and input for Zernike nomenclature is Noll: Could not identify the Zernike polynomial with this index"
                self.n, self.m = self._noll_to_nm[index]
            elif order == "ansi":
                assert (
                    index in self._ansi_to_nm
                ), "Your input for index is int and input for Zernike nomenclature is ANSI: Could not identify the Zernike polynomial with this index"
                self.n, self.m = self._ansi_to_nm[index]
        else:
            raise ValueError(
                "Could not identify your index input, we accept strings, lists and tuples only"
            )

        self.index_noll = nm_to_noll(self.n, self.m)
        self.index_ansi = nm_to_ansi(self.n, self.m)
        self.name = (
            self._ansi_names[self.index_ansi]
            if self.index_ansi < len(self._ansi_names)
            else None
        )
        self._mutable = False

    def polynomial(self, size, normed=True, outside=np.nan):

        """
        For visualization of Zernike polynomial on a disc of unit radius
        :param size: integer, Defines the shape of square grid, e.g. 256 or 512
        :param normed: boolen, Whether the Zernike polynomials are normalized, default is True
        :param outside: scalar, Outside padding of the spherical disc defined within a square grid, default is np.nan
        :return: 2D array, Zernike polynomial computed on a disc of unit radius defined within a square grid
        """

        assert np.isscalar(size) and int(size) > 0
        return self.phase(*rho_theta(int(size)), normed=normed, outside=outside)

    def phase(self, rho, theta, normed=True, outside=None):

        """
        For creation of a Zernike polynomial  with a given polar co-ordinate system
        :param rho: 2D square array,  radial axis
        :param theta: 2D square array, azimuthal axis
        :param normed: boolen, whether the Zernike polynomials are normalized, default is True
        :param outside: scalar, outside padding of the spherical disc defined within a square grid, default is None
        :return: 2D array, Zernike polynomial computed for rho and theta
        """
        assert (
            isinstance(rho, np.ndarray)
            and rho.ndim == 2
            and rho.shape[0] == rho.shape[1]
        ), "Only 2D square array for radial co-ordinate is accepted"
        assert (
            isinstance(theta, np.ndarray) and theta.shape == rho.shape
        ), "Only 2D square array for azimutha co-ordinate is accepted"
        size = rho.shape[0]
        assert np.isscalar(normed)
        assert outside is None or np.isscalar(
            outside
        ), "Only scalar constant value for outside is accepted"
        w = nm_polynomial(self.n, self.m, rho, theta, normed=bool(normed))
        if outside is not None:
            w[nm_polynomial(0, 0, rho, theta, normed=False) < 1] = outside

        return w

    def __hash__(self):
        return hash((self.n, self.m))

    def __eq__(self, other):
        return isinstance(other, Zernike) and (self.n, self.m) == (other.n, other.m)

    def __lt__(self, other):
        return self.index_ansi < other.index_ansi

    def __setattr__(self, *args):
        if self._mutable:
            super().__setattr__(*args)
        else:
            raise AttributeError("Zernike is immutable")

    def __repr__(self):
        return (
            f"Zernike(n={self.n}, m={self.m: 1}, noll={self.index_noll:2}, ansi={self.index_ansi:2}"
            + (f", name='{self.name}')" if self.name is not None else ")")
        )


class ZernikeWavefront:

    """
    Encapsulates the wavefront defined by Zernike polynomials
    :param amplitudes: dictionary, nd array, tuple or list, Amplitudes of Zernike polynomials
    :param oder: string, Zernike nomenclature, eg noll or ansi, default is noll
    """

    def __init__(self, amplitudes, order="noll"):
        amplitudes = ensure_dict(amplitudes, order)
        assert all(
            np.isscalar(a) for a in amplitudes.values()
        ), "Could not identify scalar value for amplitudes after making a dictionary"

        self.zernikes = {Zernike(j, order=order): a for j, a in amplitudes.items()}
        self.amplitudes_noll = tuple(
            dict_to_list({z.index_noll: a for z, a in self.zernikes.items()})[1:]
        )
        self.amplitudes_ansi = tuple(
            dict_to_list({z.index_ansi: a for z, a in self.zernikes.items()})
        )
        self.amplitudes_requested = tuple(
            self.zernikes[k] for k in sorted(self.zernikes.keys())
        )

    def __len__(self):
        return len(self.zernikes)

    def polynomial(self, size, normed=True, outside=np.nan):

        """
        For visualization of weighted sum of Zernike polynomials on a disc of unit radius
        :param size: integer, Defines the shape of square grid, e.g. 256 or 512
        :param normed: boolen, Whether the Zernike polynomials are normalized, default is True
        :param outside: scalar, Outside padding of the spherical disc defined within a square grid, default is np.nan
        :return: 2D array, weighted sums of Zernike polynomials computed on a disc of unit radius defined within a square grid
        """
        return np.sum(
            [
                a * z.polynomial(size=size, normed=normed, outside=outside)
                for z, a in self.zernikes.items()
            ],
            axis=0,
        )

    def phase(self, rho, theta, normed=True, outside=None):

        """
        For creation of phase defined as a weighted sum of Zernike polynomial with a given polar co-ordinate system
        :param rho: 2D square array,  radial axis
        :param theta: 2D square array, azimuthal axis
        :param normed: boolen, whether the Zernike polynomials are normalized, default is True
        :param outside: scalar, outside padding of the spherical disc defined within a square grid, default is none
        :return: 2D array, wavefront computed for rho and theta
        """
        return np.sum(
            [
                a * z.phase(rho=rho, theta=theta, normed=normed, outside=outside)
                for z, a in self.zernikes.items()
            ],
            axis=0,
        )


def random_zernike_wavefront(amplitude_ranges, order="noll", rng=None):

    """
    Creates random Zernike wavefront with random amplitudes drawn from a uniform distibution
    :param aplitude_ranges: dictionary, nd array, tuple or list, amplitude bounds
    :param oder: string, to define the Zernike nomenclature if index is an integer, eg noll or ansi, default is noll
    :param rng:
    :return: Zernike wavefront object
    """
    if rng is None:
        rng = np.random
    amplitude_ranges = ensure_dict(amplitude_ranges, order)
    assert all(
        (np.isscalar(v) and v >= 0) or (isinstance(v, (tuple, list)) and len(v) == 2)
        for v in amplitude_ranges.values()
    )
    amplitude_ranges = {
        k: ((-v, v) if np.isscalar(v) else v) for k, v in amplitude_ranges.items()
    }
    assert all(
        v[0] <= v[1] for v in amplitude_ranges.values()
    ), "Lower bound is expected to be less than the upper bound"

    return ZernikeWavefront(
        {k: rng.uniform(*v) for k, v in amplitude_ranges.items()}, order=order
    )



from faser.generators.base import Aberration
from faser.utils import generate_random_abberation
import numpy as np
import pytest


@pytest.mark.parametrize("limits", [(0, 1), (0, 0.5)])
def test_random_abberation(limits):
    abb = generate_random_abberation(limits=limits)
    assert isinstance(abb, Aberration)
    assert isinstance(abb.a1, float)
    for key, value in Aberration.__fields__.items():
        assert isinstance(getattr(abb, key), float)
        assert getattr(abb, key) >= limits[0]
        assert getattr(abb, key) <= limits[1]



from typing import Tuple
import numpy as np

# from pkg_resources import working_set  
from faser.generators.base import *
# from faser.generators.utils import polar_to_cartesian

def cart_to_polar(x, y) -> Tuple[np.ndarray, np.ndarray]:
    rho = np.sqrt(np.square(x) + np.square(y))
    # rho=rho/s.pupil_radius
    theta = np.arctan2(y, x)
    return rho, theta


def Amplitude(x, y, s: PSFConfig):
    Amp = np.exp(-(x**2 + y**2) / s.Waist**2)
    return Amp

def zernike(x: np.ndarray, y: np.ndarray, s: PSFConfig):

    temp, phi = cart_to_polar(x, y)
    rho= temp / s.r0

    Z0 = 1
    Z1 = 2 * rho * np.sin(phi)  # Tilt -vertical tilt
    Z2 = 2 * rho * np.cos(phi)  # Tip - horizontal tilt
    Z3 = np.sqrt(6) * (rho**2) * np.sin(2 * phi)  # Oblique astigmatism
    Z4 = np.sqrt(3) * (2 * rho**2 - 1)  # Defocus
    Z5 = np.sqrt(6) * (rho**2) * np.cos(2 * phi)  # Vertical astigmatism  
    Z6 = np.sqrt(8) * (rho**3) * np.sin(3 * phi)  # Vertical trefoil
    Z7 = np.sqrt(8) * (3 * rho**3 - 2 * rho) * np.sin(phi)  # Vertical coma
    Z8 = np.sqrt(8) * (3 * rho**3 - 2 * rho) * np.cos(phi)  # Horizontal coma
    Z9 = np.sqrt(8) * (rho**3) * np.cos(3 * phi)  # Oblique trefoil  
    Z12 = np.sqrt(5) * (6 * rho**4 - 6 * rho**2 + 1)  # Primary spherical   
    zer = (
        s.a0 * Z0
        + s.a1 * Z1
        + s.a2 * Z2
        + s.a3 * Z3
        + s.a4 * Z4
        + s.a5 * Z5
        + s.a6 * Z6
        + s.a7 * Z7
        + s.a8 * Z8
        + s.a9 * Z9
        + s.a12 * Z12
    )
    return zer

def Fresnel_coeff(s: PSFConfig, ca, c2a, c2at, c3a):

    t1p = 2 * s.n1 * ca / (s.n2 * ca + s.n1 * c2a)
    t2p = 2 * s.n2 * c2a / (s.n3 * c2a + s.n2 * c3a)
    r1p = (s.n2 * ca - s.n1 * c2a) / (s.n2 * ca + s.n1 * c2a)
    r2p = (s.n3 * c2a - s.n2 * c3a) / (s.n3 * c2a + s.n2 * c3a)

    t1s = 2 * s.n1 * ca / (s.n1 * ca + s.n2 * c2a)
    t2s = 2 * s.n2 * c2a / (s.n2 * c2a + s.n3 * c3a)
    r1s = (s.n1 * ca - s.n2 * c2a) / (s.n1 * ca + s.n2 * c2a)
    r2s = (s.n2 * c2a - s.n3 * c3a) / (s.n2 * c2a + s.n3 * c3a)

    beta = s.k0 * s.n2 * (s.Thickness * c2a - s.Collar * c2at)

    Tp = t2p * t1p * np.exp(1j * beta) / (1 + r1p * r2p * np.exp(2 * 1j * beta))
    Ts = t2s * t1s * np.exp(1j * beta) / (1 + r1s * r2s * np.exp(2 * 1j * beta))

    return Tp, Ts



def poisson_noise(image, seed=None):

    #Add Poisson noise to an image.


    if image.min() < 0:
        low_clip = -1.0
    else:
        low_clip = 0.0

    rng = np.random.default_rng(seed)
    # Determine unique values in image & calculate the next power of two
    vals = len(np.unique(image))
    vals = 2 ** np.ceil(np.log2(vals))

    # Ensure image is exclusively positive
    if low_clip == -1.0:
        old_max = image.max()
        image = (image + 1.0) / (old_max + 1.0)

    # Generating noise for each unique value in image.
    out = rng.poisson(image * vals) / float(vals)

    # Return image to original range if input was signed
    if low_clip == -1.0:
        out = out * (old_max + 1.0) - 1.0

    return out



def generate_intensity_profile(s: PSFConfig):

    # Discretization of the pupil plan
    x1 = np.linspace(-s.r0, s.r0, s.Nxy)
    y1 = np.linspace(-s.r0, s.r0, s.Nxy)
    X1, Y1 = np.meshgrid(x1, y1)

    rho, phi = cart_to_polar(X1, Y1)
    
    intensity= np.empty((s.Nxy, s.Nxy))*np.nan

    mask1 = rho <= s.r0
    temp=np.zeros((s.Nxy, s.Nxy))
    intensity[mask1]=temp[mask1]

    mask2 = rho <= s.r0_eff
    intensity[mask2]=Amplitude(
                    X1[mask2] - s.r0 / s.Nxy * s.Ampl_offset_x,
                    Y1[mask2] - s.r0 / s.Nxy * s.Ampl_offset_y,
                    s,
                 )
    
    return intensity


def generate_phase_mask(s: PSFConfig):

    if s.Mode == mode.DONUT_BOTTLE:  # Donut & Bottle
        raise NotImplementedError("No display Donut and Bottle")
    else:
        # Discretization of the pupil plan
        x1 = np.linspace(-s.r0, s.r0, s.Nxy)
        y1 = np.linspace(-s.r0, s.r0, s.Nxy)
        [X1, Y1] = np.meshgrid(x1, y1)

        rho, phi = cart_to_polar(X1, Y1)
        phased= np.empty((s.Nxy, s.Nxy))*np.nan

        mask1 = rho <= s.r0
        temp=np.zeros((s.Nxy, s.Nxy))
        phased[mask1]=temp[mask1]

        mask2 = rho <= s.r0_eff
        phased[mask2]=phase_mask_array(
                    X1[mask2] - s.r0 / s.Nxy * s.Mask_offset_x,
                    Y1[mask2] - s.r0 / s.Nxy * s.Mask_offset_y,
                    s,
                 )
        
    return phased



def generate_aberration(s: PSFConfig):

    # Discretization of the pupil plan
    x1 = np.linspace(-s.r0, s.r0, s.Nxy)
    y1 = np.linspace(-s.r0, s.r0, s.Nxy)
    [X1, Y1] = np.meshgrid(x1, y1)

    rho, phi = cart_to_polar(X1, Y1)
    Ab= np.empty((s.Nxy, s.Nxy))*np.nan
    
    mask1 = rho <= s.r0
    temp=np.zeros((s.Nxy, s.Nxy))
    Ab[mask1]=temp[mask1]
    
    mask2 = rho <= s.r0_eff
    Ab[mask2]=zernike(
                    X1[mask2] - s.r0 / s.Nxy * s.Aberration_offset_x,
                    Y1[mask2] - s.r0 / s.Nxy * s.Aberration_offset_y,
                    s,
                 )
        
    return Ab


# phase mask function for input array
def phase_mask_array(
    x: np.ndarray,
    y: np.ndarray,
    s: PSFConfig,
):
    rho, phi = cart_to_polar(x, y)
    
    if s.Mode == mode.GAUSSIAN:  # gaussian
        mask = np.ones(rho.shape)
    elif s.Mode == mode.DONUT:  # donut
        mask = s.VC * phi
    elif s.Mode == mode.BOTTLE:  # bottle
        cutoff_radius = s.Ring_Radius * s.r0
        mask = s.RC * np.pi * np.ones(rho.shape)
        mask[rho > cutoff_radius] = 0
    elif s.Mode == mode.LOADED: 

        print(s.loaded_phase_mask)
        raise NotImplementedError("Needs to be implemented")
        # interpolate value from



    elif s.Mode == mode.DONUT_BOTTLE:  # Donut & Bottle
        raise NotImplementedError("No display Donut and Bottle")
    else:
        raise NotImplementedError("Please use a specified Mode")
    return mask


# phase mask function for scalar input
def phase_mask(
    x: np.ndarray,
    y: np.ndarray,
    s: PSFConfig,
):
    rho, phi = cart_to_polar(x, y)
    
    if s.Mode == mode.GAUSSIAN:  # gaussian
        mask = 1
    elif s.Mode == mode.DONUT:  # donut
        mask = np.exp(1j * s.VC * phi)
    elif s.Mode == mode.BOTTLE:  # bottle
        cutoff_radius = s.Ring_Radius * s.r0
        if rho <= cutoff_radius:
            mask = np.exp(1j * s.RC * np.pi)
        else:
            mask = np.exp(1j * 0)
    elif s.Mode == mode.DONUT_BOTTLE:  # Donut & Bottle
        raise NotImplementedError("No display Donut and Bottle")
    else:
        raise NotImplementedError("Please use a specified Mode")
    return mask

def calculate_electric_field(s:PSFConfig) -> np.array:
    # Discretization of the pupil plan
    x1 = np.linspace(-s.r0, s.r0, s.Nxy)
    y1 = np.linspace(-s.r0, s.r0, s.Nxy)
    [X1, Y1] = np.meshgrid(x1, y1)

    # Discretization of the sample volumle
    x2 = np.linspace(-s.L_obs_XY, s.L_obs_XY, s.Nxy)
    y2 = np.linspace(-s.L_obs_XY, s.L_obs_XY, s.Nxy)
    z2 = np.linspace(-s.L_obs_Z, s.L_obs_Z, s.Nz)
    [X2, Y2, Z2] = np.meshgrid(x2, y2, z2+s.Dfoc/s.L_obs_Z)

    # Initialization electric field near focus
    Ex2 = 0
    Ey2 = 0
    Ez2 = 0

    #Noise = np.abs(np.random.normal(0, s.Gaussian_beam_noise, (s.Ntheta, s.Nphi)))

    theta = 0
    phi = 0
    for p in range(0, s.Ntheta):
        theta = p * s.deltatheta
        for q in range(0, s.Nphi):    #TODO check the -1
            phi = q * s.deltaphi

            ci = np.cos(phi)
            ca = np.cos(theta)
            si = np.sin(phi)
            sa = np.sin(theta)

            # refracted angles
            theta2 = np.arcsin((s.n1 / s.n2) * np.sin(theta))
            c2a = np.cos(theta2)
            theta3 = np.arcsin((s.n2 / s.n3) * np.sin(theta2))
            c3a = np.cos(theta3)
            s3a = np.sin(theta3)

            # Cartesian coordinate on pupil
            x_pup = s.WD * sa * ci
            y_pup = s.WD * sa * si

            # Rotation and projection of the pupil function
            x_pup_t = s.cg * x_pup - s.sg * s.WD
            y_pup_t = y_pup
            # Spherical coordinate
            theta_t = np.arcsin(
                np.sqrt(x_pup_t**2 + y_pup_t**2) / s.WD
            )  
            cat = np.cos(theta_t)

            # refracted tilted angles
            theta2_t = np.arcsin((s.n1 / s.n2) * np.sin(theta_t))
            c2at = np.cos(theta2_t)

            if theta_t <= s.alpha_eff:

                # Amplitude of the incident beam on the objective pupil
                Amp = Amplitude(
                    x_pup_t - s.r0 / s.Nxy * s.Ampl_offset_x,
                    y_pup_t - s.r0 / s.Nxy * s.Ampl_offset_y,
                    s,
                )
                #Amp = Amp + Noise[slice][q]

                # Phase mask on the objective pupil
                PM = phase_mask(
                    x_pup_t - s.r0 / s.Nxy * s.Mask_offset_x,
                    y_pup_t - s.r0 / s.Nxy * s.Mask_offset_y,
                    s,
                )

                # Wavefront on the objective pupil
                W = np.exp(
                    1j  # *2*np.pi
                    * zernike(
                        x_pup_t - s.r0 / s.Nxy * s.Aberration_offset_x,
                        y_pup_t - s.r0 / s.Nxy * s.Aberration_offset_y,
                        s
                        )
                    )
            else:
                Amp = 0
                PM = 1
                W = 0

            # incident beam polarization cases
            p0x = [
                np.cos(s.psi) * np.cos(s.eps) - 1j * np.sin(s.psi) * np.sin(s.eps),
                ci,
                -si,
            ]
            p0y = [
                np.sin(s.psi) * np.cos(s.eps) + 1j * np.cos(s.psi) * np.sin(s.eps),
                si,
                ci,
            ]
            p0z = 0

            # Selected incident beam polarization
            P0 = [
                [
                    p0x[s.Polarization - 1]
                ],  # indexing minus one to get corresponding polarization
                [
                    p0y[s.Polarization - 1]
                ],  # indexing minus one to get corresponding polarization
                [p0z],
            ] 

            [Tp, Ts] = Fresnel_coeff(s, ca, c2a, c2at, c3a)

            T = [   # Polarization matrix
                [
                    Tp * c3a * ci**2 + Ts * si**2,
                    Tp * si * ci * c3a - Ts * ci * si,
                    Tp * s3a * ci,
                ],
                [
                    Tp * c3a * ci * si - Ts * si * ci,
                    Tp * c3a * si**2 + Ts * ci**2,
                    Tp * s3a * si,
                ],
                [
                    -Tp * ci * s3a,
                    -Tp * s3a * si,
                    Tp * c3a,
                ],
            ]  

            # polarization in focal region
            P = np.matmul(T, P0)

            # Apodization factor
            a = np.sqrt(cat)

            # numerical calculation of electric field distribution in focal region
            propagation = (
                np.exp(
                    1j * s.k0 * s.n1 * (X2 * ci * sa + Y2 * si * sa)
                    + 1j * s.k0 * s.n3 * c3a * Z2
                )
                * s.deltaphi
                * s.deltatheta
            )

            # Aberration term from the coverslip
            Psi_coverslip = s.n3 * s.Depth * c3a - s.n1 * (s.Thickness + s.Depth) * ca
            Psi_collar = -s.n1 * s.Collar * cat  ### TO DO: ca?
            Psi_w = Psi_coverslip - Psi_collar
            Ab_wind = np.exp(1j * s.k0 * Psi_w)

            factored = sa * a * Amp * PM * W * Ab_wind * propagation

            Ex2 = Ex2 + factored * P[0, 0]
            Ey2 = Ey2 + factored * P[1, 0]
            Ez2 = Ez2 + factored * P[2, 0]
    
    return Ex2, Ey2, Ez2


def generate_psf(s: PSFConfig) -> np.ndarray:

    if s.Mode == mode.DONUT_BOTTLE:
        s.Mode=mode.DONUT   # Calculate Donut PSF
        [Ex2d, Ey2d, Ez2d] = calculate_electric_field(s)
        Ix2 = np.multiply(np.conjugate(Ex2d), Ex2d)
        Iy2 = np.multiply(np.conjugate(Ey2d), Ey2d)
        Iz2 = np.multiply(np.conjugate(Ez2d), Ez2d)
        I1d = Ix2 + Iy2 + Iz2

        s.Mode=mode.BOTTLE  # Calculated Bottle PSF
        [Ex2b, Ey2b, Ez2b] = calculate_electric_field(s)
        Ix2 = np.multiply(np.conjugate(Ex2b), Ex2b)
        Iy2 = np.multiply(np.conjugate(Ey2b), Ey2b)
        Iz2 = np.multiply(np.conjugate(Ez2b), Ez2b)
        I1b = Ix2 + Iy2 + Iz2

        I1=s.p*I1d+(1-s.p)*I1b

        s.Mode=mode.DONUT_BOTTLE

    else:
        [Ex2, Ey2, Ez2] = calculate_electric_field(s)
        Ix2 = np.multiply(np.conjugate(Ex2), Ex2)
        Iy2 = np.multiply(np.conjugate(Ey2), Ey2)
        Iz2 = np.multiply(np.conjugate(Ez2), Ez2)
        I1 = Ix2 + Iy2 + Iz2

   # I1 = I1 + np.abs(np.random.normal(0, s.Detector_gaussian_noise, I1.shape))

    if s.Normalize==normalize.YES:
        # We are only rescaling to the max, not the min
        I1 = I1 / np.max(I1)

    return np.real(np.moveaxis(I1, 2, 0))

 


from typing import Tuple
import numpy as np#
from faser.generators.base import *


def zernike(rho, theta, a: Aberration):
    Z1 = 1
    Z2 = 2 * rho * np.cos(theta)  # Tip
    Z3 = 2 * rho * np.sin(theta)  # Tilt
    Z4 = np.sqrt(3) * (2 * rho**2 - 1)  # Defocus
    Z5 = np.sqrt(6) * (rho**2) * np.cos(2 * theta)  # Astigmatisme
    Z6 = np.sqrt(6) * (rho**2) * np.sin(2 * theta)  # Astigmatisme
    Z7 = np.sqrt(8) * (3 * rho**3 - 2 * rho) * np.cos(theta)  # coma
    Z8 = np.sqrt(8) * (3 * rho**3 - 2 * rho) * np.sin(theta)  # coma
    Z9 = np.sqrt(8) * (rho**3) * np.cos(3 * theta)  # Trefoil
    Z10 = np.sqrt(8) * (rho**3) * np.sin(3 * theta)  # Trefoil
    Z11 = np.sqrt(5) * (6 * rho**4 - 6 * rho**2 + 1)  # Spherical
    zer = (
        a.a1 * Z1
        + a.a2 * Z2
        + a.a3 * Z3
        + a.a4 * Z4
        + a.a5 * Z5
        + a.a6 * Z6
        + a.a7 * Z7
        + a.a8 * Z8
        + a.a9 * Z9
        + a.a10 * Z10
        + a.a11 * Z11
    )
    return zer


def poisson_noise(image, seed=None):
    """
    Add Poisson noise to an image.
    """

    if image.min() < 0:
        low_clip = -1.0
    else:
        low_clip = 0.0

    rng = np.random.default_rng(seed)
    # Determine unique values in image & calculate the next power of two
    vals = len(np.unique(image))
    vals = 2 ** np.ceil(np.log2(vals))

    # Ensure image is exclusively positive
    if low_clip == -1.0:
        old_max = image.max()
        image = (image + 1.0) / (old_max + 1.0)

    # Generating noise for each unique value in image.
    out = rng.poisson(image * vals) / float(vals)

    # Return image to original range if input was signed
    if low_clip == -1.0:
        out = out * (old_max + 1.0) - 1.0

    return out


# phase mask function
def phase_mask(
    rho: np.ndarray,
    theta: np.ndarray,
    cutoff_radius: float,
    vortex_charge: float,
    ring_charge: float,
    mode: Mode,
):
    if mode == Mode.GAUSSIAN:  # guassian
        mask = 1
    elif mode == Mode.DONUT:  # donut
        mask = np.exp(1j * vortex_charge * theta)
    elif mode == Mode.BOTTLE:  # bottleMo
        if rho < cutoff_radius:
            mask = np.exp(1j * ring_charge * np.pi)
        else:
            mask = np.exp(1j * 0)
    else:
        raise NotImplementedError("Please use a specified Mode")
    return mask


def cart_to_polar(x, y) -> Tuple[np.ndarray, np.ndarray]:
    rho = np.sqrt(np.square(x) + np.square(y))
    theta = np.arctan2(y, x)
    return rho, theta


def pol_to_cart(rho, phi):
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return (x, y)


def generate_psf(s: PSFConfig) -> np.ndarray:

    # Calulcated Parameters
    wavenumber = 2 * np.pi / s.wavelength  # wavenumber

    focusing_angle = np.arcsin(
        s.numerical_aperature / s.refractive_index_immersion
    )  # maximum focusing angle of the objective

    pupil_radius = s.working_distance * np.tan(focusing_angle)  # radius of the pupil

    # Sample Space
    x1 = np.linspace(-pupil_radius, pupil_radius, s.Nx)
    y1 = np.linspace(-pupil_radius, pupil_radius, s.Ny)
    [X1, Y1] = np.meshgrid(x1, y1)

    x2 = np.linspace(-s.LfocalX, s.LfocalX, s.Nx)
    y2 = np.linspace(-s.LfocalY, s.LfocalY, s.Ny)
    z2 = np.linspace(-s.LfocalZ, s.LfocalZ, s.Nz)
    [X2, Y2, Z2] = np.meshgrid(x2, y2, z2)  # TODO: Needs to be propüerly constructed

    ## incident beam polarization cases
    p0x = [1, 0, 1 / np.sqrt(2), 1j / np.sqrt(2)]
    p0y = [0, 1, 1j / np.sqrt(2), 1 / np.sqrt(2)]
    p0z = 0

    P0 = [
        [
            p0x[s.polarization - 1]
        ],  # indexing minus one to get corresponding polarization
        [
            p0y[s.polarization - 1]
        ],  # indexing minus one to get corresponding polarization
        [p0z],
    ]  #

    # TODO: Bring to loop to allow other polaraization

    # Step of integral
    deltatheta = focusing_angle / (s.Ntheta - 1)
    deltaphi = 2 * np.pi / s.Nphi

    # Initialization
    Ex2 = 0  # Ex?component in focal
    Ey2 = 0  # Ey?component in focal
    Ez2 = 0

    Noise = np.abs(np.random.normal(0, s.gaussian_beam_noise, (s.Ntheta, s.Nphi)))

    theta = 0
    phi = 0
    for slice in range(0, s.Ntheta):
        theta = slice * deltatheta
        for q in range(0, s.Nphi):
            phi = q * deltaphi

            ci = np.cos(phi)
            ca = np.cos(theta)
            si = np.sin(phi)
            sa = np.sin(theta)

            T = [
                [
                    1 + (ci**2) * (ca - 1),
                    si * ci * (ca - 1),
                    sa * ci,
                ],
                [
                    si * ci * (ca - 1),
                    ca * (si**2) + (ci**2),
                    sa * si,
                ],
                [
                    -sa * ci,
                    -sa * si,
                    ca,
                ],
            ]  # Pola matrix

            # polarization in focal region
            P = np.matmul(T, P0)
            # Cylindrical coordinates on pupil

            x_pup = s.working_distance * sa * ci
            y_pup = s.working_distance * sa * si

            # Cylindrical coordinates on pupil
            rho_pupil_polar, theta_pupil_polar = cart_to_polar(x_pup, y_pup)
            rho_amp_polar, theta_amp_polar = cart_to_polar(
                x_pup - pupil_radius / s.Nx * s.ampl_offsetX,
                y_pup - pupil_radius / s.Ny * s.ampl_offsetY,
            )
            rho_mask_polar, theta_mask_polar = cart_to_polar(
                x_pup - pupil_radius / s.Nx * s.mask_offsetX,
                y_pup - pupil_radius / s.Ny * s.mask_offsetY,
            )
            rho_ab_polar, theta_ab_polar = cart_to_polar(
                x_pup - pupil_radius / s.Nx * s.aberration_offsetX,
                y_pup - pupil_radius / s.Ny * s.aberration_offsetY,
            )
            # Apodization factor
            a = np.sqrt(ca)
            # Incident intensity profile
            Bi = np.exp(-np.square(rho_amp_polar) / np.square(s.beam_waist))
            Bi = Bi + Noise[slice][q]
            # Phase mask
            PM = phase_mask(
                rho_mask_polar,
                theta_mask_polar,
                s.unit_phase_radius * pupil_radius,
                s.vortex_charge,
                s.ring_charge,
                s.mode,
            )
            # Wavefront
            W = np.exp(
                1j
                * np.pi
                * zernike(rho_ab_polar / pupil_radius, theta_ab_polar, s.aberration)
            )

            # numerical calculation of field distribution in focal region

            propagation = (
                np.exp(
                    1j
                    * wavenumber
                    * s.refractive_index_immersion
                    * (X2 * ci * sa + Y2 * si * sa + Z2 * ca)
                )
                * deltaphi
                * deltatheta
            )

            factored = sa * a * Bi * PM * W * propagation

            Ex2 = Ex2 + factored * P[0, 0]
            Ey2 = Ey2 + factored * P[1, 0]
            Ez2 = Ez2 + factored * P[2, 0]

    Ix2 = np.multiply(np.conjugate(Ex2), Ex2)
    Iy2 = np.multiply(np.conjugate(Ey2), Ey2)
    Iz2 = np.multiply(np.conjugate(Ez2), Ez2)
    I1 = Ix2 + Iy2 + Iz2
    I1 = np.real(I1)

    # Add Poisson noise to I1
    if s.add_detector_poisson_noise:
        I1 = poisson_noise(I1)

    if s.rescale:
        # We are only rescaling to the max, not the min
        I1 = I1 / np.max(I1)

    return np.moveaxis(I1, 2, 0)



import numpy as np
from scipy.interpolate import griddata

def polar_to_cartesian(polar_image):
    """
    Convert an image in polar coordinates to Cartesian coordinates.
    :param polar_image: A 2D numpy array where the first dimension is the radius and the second is the angle.
    :return: A 2D numpy array in Cartesian coordinates.
    """
    max_radius, num_angles = polar_image.shape
    
    # Generate r and theta arrays
    r = np.linspace(0, max_radius, max_radius)
    theta = np.linspace(0, 2*np.pi, num_angles)
    R, Theta = np.meshgrid(r, theta)

    # Convert polar (R, Theta) to Cartesian (X, Y)
    X, Y = R * np.cos(Theta), R * np.sin(Theta)

    # Flatten X, Y, and polar_image for griddata interpolation
    x = X.ravel()
    y = Y.ravel()
    z = polar_image.ravel()

    # Define a grid of Cartesian coordinates
    xi, yi = np.linspace(x.min(), x.max(), polar_image.shape[1]), np.linspace(y.min(), y.max(), polar_image.shape[0])
    xi, yi = np.meshgrid(xi, yi)

    # Interpolate using griddata
    cartesian_image = griddata((x, y), z, (xi, yi))

    return np.angle(cartesian_image)


def polar_phase_mask(num_radii, num_angles):
    r = np.linspace(0, 1, num_radii)
    theta = np.linspace(0, 2*np.pi, num_angles)
    R, Theta = np.meshgrid(r, theta)
    phase = (2 * np.pi * R) % (2 * np.pi)
    return phase


from typing import Tuple
from .generators.base import Aberration
import numpy as np


def generate_random_abberation(
    limits: Tuple[float, float], limit_map=None
) -> Aberration:
    if not limit_map:
        limit_map = {
            key: (limits[0], limits[1]) for key, value in Aberration.__fields__.items()
        }

    return Aberration(
        **{key: float(np.random.uniform(*value)) for key, value in limit_map.items()}
    )



import sys
import typing
from PyQt5.QtWidgets import QWidget
import matplotlib
import numpy as np
matplotlib.use("Qt5Agg")

from PyQt5 import QtCore, QtWidgets

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
from matplotlib.cm import coolwarm, viridis, turbo

class MplCanvas(FigureCanvasQTAgg):
    def __init__(self, parent=None, width=5, height=4, dpi=300):
        self.fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = self.fig.add_subplot(111)
        super(MplCanvas, self).__init__(self.fig)


class MatplotlibDialog(QtWidgets.QDialog):

    def __init__(self, title, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setWindowTitle(title)
        self.title=title

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setSpacing(1)
        self.setLayout(self.layout)

        self.again = QtWidgets.QPushButton("Popup")

        self.canvas = MplCanvas(self, width=5, height=5, dpi=100)
        self.canvas.axes.set_aspect('auto')
        self.layout.addWidget(self.canvas)
        
        self.layout.addWidget(self.again)
        self.again.clicked.connect(self.show_another)
        self.another = None

    def update(self, wavefront, new_title):
        self.data = wavefront
        self.canvas.axes.imshow(wavefront)
        self.canvas.axes.set_title(new_title)
        self.canvas.draw()

    def show_another(self) -> None:
        self.another = self.__class__(parent=self, title="Copy of " + self.title)
        self.another.update(self.data, "Copy of " + self.title)
        self.another.show()


class WavefrontDialog(MatplotlibDialog):

    def __init__(self, title, *args, **kwargs) -> None:
        super().__init__(title, *args, **kwargs)
        self.colorbar = None


    def update(self, wavefront, new_title):
        if self.colorbar is not None:
            self.colorbar.remove()
        self.canvas.axes.set_title(new_title)
        self.data = wavefront
        image = self.canvas.axes.imshow(wavefront, cmap=coolwarm)
        image.set_clim(-np.pi, np.pi)
        self.colorbar = self.canvas.fig.colorbar(image, ax=self.canvas.axes, orientation='vertical')
        self.canvas.draw()
    

class BeamDialog(MatplotlibDialog):

    def __init__(self, title, *args, **kwargs) -> None:
        super().__init__(title, *args, **kwargs)
        self.colorbar = None


    def update(self, wavefront, new_title):
        if self.colorbar is not None:
            self.colorbar.remove()
        self.canvas.axes.set_title(new_title)
        self.data = wavefront
        image = self.canvas.axes.imshow(wavefront, cmap=viridis)
        image.set_clim(0, 1)
        self.colorbar = self.canvas.fig.colorbar(image, ax=self.canvas.axes, orientation='vertical')
        self.canvas.draw()
    

class PhaseMaskDialog(MatplotlibDialog):

    def __init__(self, title, *args, **kwargs) -> None:
        super().__init__(title, *args, **kwargs)
        self.colorbar = None


    def update(self, wavefront, new_title):
        if self.colorbar is not None:
            self.colorbar.remove()
        self.canvas.axes.set_title(new_title)
        self.data = wavefront
        image = self.canvas.axes.imshow(wavefront, cmap=turbo)
        image.set_clim(-np.pi, np.pi)
        self.colorbar = self.canvas.fig.colorbar(image, ax=self.canvas.axes, orientation='vertical')
        self.canvas.draw()


from faser.generators.base import PSFGenerator
from faser.generators.scalar.phasenet.generator import PsfGenerator3D
from .wavefront import ZernikeWavefront


class PhaseNetPSFGenerator(PSFGenerator):
    def generate(self):

        psf = PsfGenerator3D(
            psf_shape=(self.config.Nx, self.config.Ny, self.config.Nz),
            units=(0.1, 0.1, 0.1),  # TODO: Figure out how to solve?
            na_detection=self.config.numerical_aperature,
            lam_detection=self.config.wavelength * 10e5,  # meter to microns
            n=self.config.refractive_index,
        )

        print(self.config.wavelength * 10e5)

        # Aberration to ansi
        params = {
            int(key[1:]): value for key, value in self.config.aberration.dict().items()
        }

        wf = ZernikeWavefront(params, order="ansi")

        return psf.incoherent_psf(wf)



#%%
import numpy as np
from faser.generators.base import PSFConfig, Mode

from faser.generators.vectorial.stephane import generate_psf

#%%
config = PSFConfig(Nx=32, Ny=32, Nz=32, mode=Mode.DONUT)


h1 = generate_psf(config)
print(h1.max())
h1.shape

# %%
h1.dtype
# %%
h1
# %%



import napari
import numpy as np
from magicgui import magicgui
# from faser.generators.base import PSFConfig, mode, window
from faser.generators.vectorial.stephane.tilted_coverslip import generate_psf
import numpy as np

# import napari
# from qtpy.QtWidgets import QHBoxLayout, QPushButton, QWidget
from scipy import ndimage
from skimage.transform import resize

# from faser.generators.base import Aberration, Mode, WindowType, Polarization, PSFConfig
# from faser.generators.vectorial.stephane.tilted_coverslip import generate_psf

slider = {"widget_type": "FloatSlider", "min": -1, "max": 1, "step": 0.05}
tilt_slider = {"widget_type": "FloatSlider", "min": 0, "max": 90, "step": 0.05}
detector_slider = {"widget_type": "FloatSlider", "min": 0, "max": 1, "step": 0.05}
focal_slider = {"widget_type": "Slider", "min": 1, "max": 10, "step": 1}
beam_slider = {"widget_type": "FloatSlider", "min": 0.5, "max": 50, "step": 0.5}

viewer = None


""" @magicgui(
    call_button="Generate",
    LfocalXY=focal_slider,
    LfocalZ=focal_slider,
    piston=slider,
    tip=slider,
    tilt=slider,
    defocus=slider,
    astigmatism_v=slider,
    astigmatism_h=slider,
    coma_v=slider,
    coma_h=slider,
    trefoil_v=slider,
    trefoil_h=slider,
    spherical=slider,
    tilt_angle=tilt_slider,
    gaussian_beam_noise=detector_slider,
    detector_gaussian_noise=detector_slider,
) """
def generate_psf_gui ():
    """viewer: napari.Viewer,
    Nx=31,  # discretization of image plane
    Ny=31,
    Nz=31,
    LfocalXY=2,  # observation scale X and Y
    LfocalZ=4,  # observation scale Z
    Ntheta=31,  # Integration steps
    Nphi=31,
    # Optical aberrations
    piston=0.0,
    tip=0.0,
    tilt=0.0,
    defocus=0.0,
    astigmatism_v=0.0,
    astigmatism_h=0.0,
    coma_v=0.0,
    coma_h=0.0,
    trefoil_v=0.0,
    trefoil_h=0.0,
    spherical=0.0,
    tilt_angle=0.0,
    gaussian_beam_noise=0.0,
    detector_gaussian_noise=0.0,
    add_detector_poisson_noise=False,
    Normalize=True,
    # Phase profile
    Mode: mode = mode.GAUSSIAN,
    Window: window = window.NO,
    psi=0,
    eps=45,
    """
    """    config = PSFConfig(
        a1=piston,
        a2=tip,
        a3=tilt,
        a4=defocus,
        a5=astigmatism_v,
        a6=astigmatism_h,
        a7=coma_v,
        a8=coma_h,
        a9=trefoil_v,
        a10=trefoil_h,
        a11=spherical,
        Nx=Nx,
        Ny=Ny,
        Nz=Nz,
        Ntheta=Ntheta,
        Nphi=Nphi,
        Mode=Mode,
        Polarization=Polarization,
        Window=Window,
        gaussian_beam_noise=gaussian_beam_noise,
        detector_gaussian_noise=detector_gaussian_noise,
        add_detector_poisson_noise=add_detector_poisson_noise,
        LfocalX=LfocalXY * 1e-6,
        LfocalY=LfocalXY * 1e-6,  # observation scale Y
        LfocalZ=LfocalZ * 1e-6,
        Normalize=Normalize,
        psi_degree=psi,
        eps_degree=eps,
        tilt_angle_degree=tilt_angle,
    )
 """
    #psf = generate_psf(config)
    #print(psf.max())
    #return viewer.add_image(
    #    psf,
    #    name=f"PSF {config.mode.name} {config} ",
    #    metadata={"is_psf": True, "config": config},
    #    colormap="viridis",
    #)


@magicgui(
    call_button="Effective PSF",
    I_sat=slider,
)
def make_effective_gui(viewer: napari.Viewer, I_sat=0.1):

    gaussian_layers = (
        layer for layer in viewer.layers.selection if layer.metadata.get("is_psf", True)
    )

    psf_layer_one = next(gaussian_layers)
    psf_layer_two = next(gaussian_layers)
    new_psf = np.multiply(psf_layer_one.data, np.exp(-psf_layer_two.data / I_sat))

    return viewer.add_image(
        new_psf,
        name=f"Combined PSF {psf_layer_one.name} {psf_layer_two.name}",
        metadata={"is_psf": True},
    )


""" @magicgui(
    call_button="Convolve Image",
) """
def convolve_image_gui(viewer: napari.Viewer, resize_psf=0):

    psf_layer = next(
        layer
        for layer in viewer.layers.selection
        if layer.metadata.get("is_psf", False)
    )
    image_layer = next(
        layer
        for layer in viewer.layers.selection
        if not layer.metadata.get("is_psf", False)
    )

    image_data = image_layer.data
    psf_data = psf_layer.data

    if image_data.ndim == 2:
        psf_data = psf_data[psf_data.shape[0] // 2, :, :]

        con = ndimage.convolve(
            image_data, psf_data, mode="constant", cval=0.0, origin=0
        )

    if resize_psf > 0:
        psf_data = resize(psf_data, (resize_psf,) * psf_data.ndim)

    con = ndimage.convolve(image_data, psf_data, mode="constant", cval=0.0, origin=0)

    return viewer.add_image(
        con.squeeze(),
        name=f"Convoled {image_layer.name} with {psf_layer.name}",
    )


""" @magicgui(
    call_button="Generate Space",
) """
def generate_space(
    viewer: napari.Viewer, x_size=100, y_size=100, z_size=20, dots: int = 50
):

    x = np.random.randint(0, x_size, size=(dots))
    y = np.random.randint(0, y_size, size=(dots))
    z = np.random.randint(0, z_size, size=(dots))

    M = np.zeros((z_size, x_size, y_size))
    for p in zip(z, x, y):
        M[p] = 1

    viewer.add_image(M, name="Space")


""" @magicgui(
    call_button="Generate Space",
) """
def calculate_fwhm(viewer: napari.Viewer):

    psf_layer = next(
        layer
        for layer in viewer.layers.selection
        if layer.metadata.get("is_psf", False)
    )

    psf_data = psf_layer.data
    # caluclate max coordinate
    max_coord = np.argmax(psf_data, axis=None)
    viewer.add_image(M, name="Space")


from faser.napari.gui import generate_psf_gui, convolve_image_gui, make_effective_gui, generate_space
# from skimage import data
import napari
# import numpy as np
import argparse
# from scipy.sparse import random
from faser.napari.widgets.main_widget import MainWidget
from faser.napari.widgets.helper_widget import HelperWidget
# from faser.napari.widgets.tob_bar import TopBar


def main(**kwargs):
    viewer = napari.Viewer()

    main = MainWidget(viewer)
    helper = HelperWidget(viewer)
    viewer.window.add_dock_widget(main, area="right", name="PSF")
    viewer.window.add_dock_widget(helper, area="right", name="Helper")


    napari.run()


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    args = parser.parse_args()

    main()
